#ifndef COMMAND_LINE_PARSER_INCLUDED
#define COMMAND_LINE_PARSER_INCLUDED

#include "Util/CommandLine.h"
#include<string>
using std::string;
#include <map>
using std::map;

class File;
/**
 * This is a command line parsing utility.  It allows command line options to
 * be bound to variables with a program.  There are two basic usage models.  In
 * the first, you can just declare a CommandLineParser and use AddOption() to
 * configure it.  In the second, you use inheritance to give command line
 * parsing ability to another object.  This can be useful because it allows the
 * options for a particular object (a Component, for instance) to be completely
 * contained.
 *
 * The class shoud be used in conjunction with CommandLine and
 * CommandLinePareserSet.  The former is just a c++ version of the normal C
 * command line, and the latter is a collection of parsers that can be applied
 * all at once to a command line.
 */
class CommandLineParser {
 public:

     /**
      * This is a bit of macro kung-fu, but it is not that bad.  The code that
      * gets produced is quite comprehensible, so if you need to debug/change
      * it start by running g++ -E to see what is going on.  Really, it's not
      * that bad.
      */
#define STRING_TO_C_STRING(a) ((a).c_str())
#define BOOL_TO_STRING(a)     ((a) ? "true" : "false")
#define IDENTITY(a)           (a)

#define VECTOR(a) VECTOR_OF_##a     
#define PARSABLE_TYPES													\
     PARSE_TYPE(INT8,    signed char  , strtoll (s.c_str(), NULL, 0), 1, "%hhd", IDENTITY)				\
     PARSE_TYPE(UINT8,   unsigned char, strtoull(s.c_str(), NULL, 0), 1, "%hhu", IDENTITY)				\
     PARSE_TYPE(UINT32,  uint32       , strtoull(s.c_str(), NULL, 0), 1, "%u",   IDENTITY)				\
     PARSE_TYPE(INT32,   int32        , strtoll (s.c_str(), NULL, 0), 1, "%d",   IDENTITY)				\
     PARSE_TYPE(UINT64,  uint64       , strtoull(s.c_str(), NULL, 0), 1, "%lld", IDENTITY)				\
     PARSE_TYPE(INT64,   int64        , strtoll (s.c_str(), NULL, 0), 1, "%llu", IDENTITY)				\
     PARSE_TYPE(DOUBLE,  double       , strtod  (s.c_str(), NULL),    1, "%f",   IDENTITY)				\
     PARSE_TYPE(FLOAT,   float        , strtod  (s.c_str(), NULL),    1, "%f",   IDENTITY)				\
     PARSE_TYPE(STRING,  string       , s                           , 1, "%s",   STRING_TO_C_STRING)			\
     PARSE_TYPE(CSTRING, char *       , strdup(s.c_str())           , 1, "%s",   IDENTITY)				\
     PARSE_TYPE(BOOL,    bool         , (strcasecmp(s.c_str(), "true") == 0 ? true : false) , 1, "%s", BOOL_TO_STRING)	\

     enum OptionType {
	  INVALID,
#define PARSE_TYPE(TYPE, type, parse, eat, format, conv) TYPE, VECTOR(TYPE),
	  PARSABLE_TYPES
#undef PARSE_TYPE
	  
     };
     
 protected:
     class Option : public Printable{
     public:
	  OptionType type;
	  string name;
	  string fullName;
	  string description;
	  void *storage;

	  Option() : name(""), fullName(""), description("") {
	       type = INVALID; 
	       storage = NULL;
	  };
	  
	  Option(OptionType aType, char *aName, char *desc, void *s) : name(aName), fullName(aName), description(desc)
	  {
	       type = aType;
	       storage = s;
	  }

	  static bool ParseItem(string s, OptionType type, void *storage);
	  static bool ParseList(string s, OptionType type, void *l);
	  bool Parse(string value);
	  bool Parse(CommandLine *cmdLine, int index);
	  string GetName() {return fullName;}
	  const char* ToString(string &b);
     };

public:

     vector<Option> options;
     vector<CommandLineParser *> subParsers;
     string stopOption;
     string loadFileOption;

#define PARSE_TYPE(TYPE, type, parse, eat, format, conv) 						\
     void AddOption(type *v, char *option, char *description,  type defaultValue) { 			\
	  Option d(TYPE, option, description, v);							\
	  options.push_back(d);										\
          *v = defaultValue;										\
     }													\
													\
     void AddOption(vector<type> *v, char *option, char *description, vector<type> defaultValue) { 	\
	  Option d(VECTOR(TYPE), option, description,v);						\
	  options.push_back(d);										\
          *v = defaultValue;										\
     }													\
     void AddOption(type *v, char *option, char *description) { 					\
	  Option d(TYPE, option, description,v);							\
	  options.push_back(d);										\
     }													\
													\
     void AddOption(vector<type> *v, char *option, char *description) { 				\
	  Option d(VECTOR(TYPE), option, description,v);						\
	  options.push_back(d);										\
     }

     PARSABLE_TYPES;
#undef PARSE_TYPE

 private:


 public:
     
     CommandLineParser() : stopOption("--"), loadFileOption("-cmdLineFile") {	  
     }

     void SetStopOption(char *option) {
	  stopOption = (string)option;
     }
     
     enum  CmdLineParseMode {MODE_STRICT, MODE_NOT_STRICT, MODE_INVALID};

     bool ParseCommandLine(CommandLine *cmdLine, CmdLineParseMode mode = MODE_STRICT);
     void CollectAllOptions(vector<Option> &allOptions);
     bool CheckCommandLine(CommandLine *cmdLine);
     void AddParser(CommandLineParser *p) {     
	  subParsers.push_back(p);
     };
     void PrintUsage(File &f);
     bool ParseFile(const string & fileName, 
		    vector<Option> & allOptions, 
		    CmdLineParseMode mode);
	  
     char **GetUnparsedArgv(CommandLine *cmdLine);     
     virtual void GetUsageString(string &buf);
     void PrintOptionsAndValues(File &f);
     virtual ~CommandLineParser(){};
     void AddPrefix(const string &prefix);

     string MunchPattern(const string &pattern);
     bool Matches(const string & pattern, Option &option);
     bool DoMatches(CommandLine *cmdLine, uint index, vector<Option> & allOptions);
     bool DoMatches(const string& key, const string &value, vector<Option> & allOptions);

};

#endif
