#include "File.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>

#include <sys/time.h>
#include <time.h>


#ifdef Darwin
const int    File::READ_WRITE_FLAGS = O_CREAT | O_RDWR;
const int    File::WRITE_FLAGS = O_CREAT | O_WRONLY | O_TRUNC;
const int    File::READ_FLAGS = O_RDONLY;
#else
#include <linux/unistd.h>
const int    File::READ_WRITE_FLAGS = O_CREAT | O_RDWR | O_LARGEFILE;
const int    File::READ_FLAGS = O_RDONLY | O_LARGEFILE;
const int    File::WRITE_FLAGS = O_CREAT | O_WRONLY | O_LARGEFILE | O_TRUNC;
_syscall5(int, _llseek, uint, fd, ulong, hi, ulong, lo, loff_t *, res, uint, wh);
#endif

const mode_t File::DEFAULT_MODE  = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;

File File::Stdout(stdout);
File File::Stderr(stderr);
File File::Stdin(stdin);

int File::tmpFileCount = 0;

#define CRASH_OR_FAIL(err) 			\
do { 						\
     myErrno = err; 				\
     if (crashOnError) {			\
	  Crash(); 				\
     } else {					\
	  return ERROR;       			\
     }						\
} while(0)
	  
File::File (bool aCrashOnError) {
     myErrno = 0;
     fDes = -1;
     fFile = NULL;
     isTmpFile = false;
     crashOnError = aCrashOnError;
}

File::File (FILE *f, bool aCrashOnError) {
     SetFile(f);
     isTmpFile = false;
     crashOnError = aCrashOnError;
}

File::File (int fd, char *mode, bool aCrashOnError) {
     SetFile(fd, mode);
     isTmpFile = false;
     crashOnError = aCrashOnError;
}

void File::Crash()
{
     fprintf(stderr, "%s\n", strerror(myErrno));
     abort();
}

File::FileStatus File::Open(const char *name, const int flags, const mode_t mode)
{
     myErrno = 0;

     fDes = open(name, flags, mode);
     if(fDes == -1) {
	  CRASH_OR_FAIL(errno);
     }
     
     /*
      * I don't how to find out which mode should be passed fdopen, so we try
      * everything until we find something that works or give up.
      */
     fFile = fdopen(fDes, "r+");
     if (fFile == NULL && errno == EINVAL) {
	  fFile = fdopen(fDes, "r");

	  if (fFile == NULL && errno == EINVAL) {
	       fFile = fdopen(fDes, "w");

	       if(fFile == NULL) {
		    CRASH_OR_FAIL(errno);
	       }
	
	  } else if (fFile == NULL) {
	       CRASH_OR_FAIL(errno);
	  }
	  
     } else if (fFile == NULL) {
	  CRASH_OR_FAIL(errno);
     }
     

     return SUCCESS;
}

File::FileStatus File::SetFile(FILE *f)
{
     myErrno = 0;
     fDes = fileno(f);
     if (fDes == -1) {
	  CRASH_OR_FAIL(errno);
     }
     fFile = f;
     
     return SUCCESS;
}

File::FileStatus File::SetFile(int fd, char *mode)
{
     myErrno = 0;
     fFile = fdopen(fd, mode);
     if (fFile == NULL) {
	  CRASH_OR_FAIL(errno);
     }
     fDes = fd;
     return SUCCESS;
}

File::FileStatus File::Open(const string &name, const int flags, const mode_t mode)
{
     return Open(name.c_str(), flags);
}    

File::FileStatus File::Write(const void *buf, const size_t n, size_t *written)
{
     return WriteN(buf, 1, n, written);
}    

File::FileStatus File::WriteN(const void *buf, const size_t size, const size_t count, size_t *written)
{
     ssize_t c;
     myErrno = 0;
     
     c = write(fDes, buf, size * count);
     
     if (c == -1) {
	  CRASH_OR_FAIL(errno);
     }
     
     if (written != NULL) {
	  *written = c;
     }
     return SUCCESS;
}

File::FileStatus File::WriteFileContents(File &f, loff_t bytes, loff_t * wr)
{
     char buf[4096];
     size_t totalWritten = 0;
     size_t remainToWrite = bytes;
     size_t w = 0;
     FileStatus r;

     while(true) {
	  size_t toRead = remainToWrite > 4096 ? 4096 : remainToWrite;
	  r = f.Read(buf, toRead, &w);

	  if (r == ERROR) {
	       CRASH_OR_FAIL(errno);
	  } else {
	       if (Write(buf, w) != SUCCESS) {
		    CRASH_OR_FAIL(errno);
	       }	 
	       remainToWrite -= w;
	       totalWritten += w;
	       if (remainToWrite == 0 ||
		   r == END_OF_FILE) {
		    break;
	       }
	  } 
     }

     if (wr != NULL) {
	  *wr = totalWritten;
     }

     return r;
}

File::FileStatus File::Read(void *buf, const size_t n, size_t * read)
{
     return ReadN(buf, 1, n, read);
}    

File::FileStatus File::ReadN(void *buf, const size_t size, const size_t count, size_t *readBytes)
{
     ssize_t c;
     myErrno = 0;
     
     c = read(fDes, buf, size * count);

     if (readBytes != NULL) {
	  *readBytes = c/size;
     }
     
     if (c == -1) {
	  CRASH_OR_FAIL(errno);
     } else if ((size_t)c != size * count) {
	  return END_OF_FILE;
     }

     
     return SUCCESS;
}

File::FileStatus File::Read(string &buf, const char delimiter)
{
     char c;
     size_t readBytes = 0;
     buf.clear();
     FileStatus r;
     while((r = ReadN(&c, sizeof(char), 1, &readBytes)) == SUCCESS) {
	  buf += c;

	  if (c == delimiter) {
	       break;
	  }
     }
     return r;
}

File::FileStatus File::Read(char &b)
{
     return ReadN(&b, sizeof(char), 1, NULL);
}    

File::FileStatus File::Write(const string &buf)
{
     return WriteN(buf.c_str(), buf.size(), 1);
}

File::FileStatus File::Write(const char c)
{
     return WriteN(&c, sizeof(char), 1);
}

int File::Print(const char*fmt, va_list ap)
{
     return vfprintf(fFile, fmt, ap);
}

int File::Print(const char*fmt, ...) 
{
     int ret;
     va_list args;
     va_start(args,fmt);
     ret = vfprintf(fFile, fmt, args);
     va_end(args);
     return ret;
}

File::FileStatus File::Seek(const int64 offset, const int whence) 
{

     int ret;
     loff_t crap;
#ifdef Linux
     ret = _llseek(fDes, (offset >> 32) & ((1ll << 32) - 1), offset & ((1ll << 32) - 1), &crap, whence);
#else
     assert(offset < INT_MAX);
     ret = lseek(fDes, offset & ((1ll << 32) - 1), whence);
#endif
     if (ret == -1) {
	  CRASH_OR_FAIL(errno);
     }
     return SUCCESS;
}

File::FileStatus File::Tell(loff_t &loc)
{
#ifdef Linux
     int ret;
     ret = _llseek(fDes, 0, 0, &loc, SEEK_CUR);
     if (ret == -1) {
	  CRASH_OR_FAIL(errno);
     }
#else
     errno = 0;
     loc = lseek(fDes, 0, SEEK_CUR);
     if (errno != 0) {
	  CRASH_OR_FAIL(errno);
     }
#endif

     return SUCCESS;
}

FILE *File::GetFILE()
{
     return fFile;
}

int File::GetFD()
{
     return fDes;
}

File::FileStatus File::Rewind()
{
     return Seek(0, SEEK_SET);
}

File::FileStatus File::Close()
{
     myErrno = 0;
     int r;
     if (fFile != NULL) {
	  r = fclose(fFile);
	  if(r == EOF) {
	       CRASH_OR_FAIL(errno);
	  }
	  fDes = -1;
     } else {
	  if (fDes != -1) {
	       r = close(fDes);

	       if (r == -1) {
		    CRASH_OR_FAIL(errno);
	       }
	       fDes = -1;
	       fFile = NULL;
	  }
     }
     return SUCCESS;
}

File::FileStatus File::DestroyTmpFile()
{
     int ret;
     assert(isTmpFile);
     FileStatus r = Close();
     
     if (r == SUCCESS) {
	  ret = unlink(tmpFileName.c_str());
     
	  if (ret != 0) {
	       CRASH_OR_FAIL(errno);
	  }
     } 
     isTmpFile = false;
     return r;
}

File::FileStatus File::OpenTmpFile(char *tmpDir, int flags)
{
     char buf[1000];
     string path(tmpDir);
     struct timeval tv;
     int ret;

     ret = gettimeofday(&tv, NULL);
     if (ret != 0) {
	  CRASH_OR_FAIL(errno);
     }
     sprintf(buf, "%d-%lu-%d.tmp", getpid(), tv.tv_sec, tmpFileCount++);

     path += buf;
     tmpFileName = path;
     isTmpFile = true;
     return Open(path, flags);
}
