#ifndef FILE_INCLUDED
#define FILE_INCLUDED

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <assert.h>

#include"Types.h"
#include <stdarg.h>
#include<string>
using std::string;
#include <stdio.h>

#ifdef Darwin
typedef off_t loff_t;
#endif

#define CRASH_OR_FAIL(err) 			\
do { 						\
     myErrno = err; 				\
     if (crashOnError) {			\
	  Crash(); 				\
     } else {					\
	  return ERROR;       			\
     }						\
} while(0)
	  
class File {
protected:
     int fDes;
     FILE *fFile;
     int myErrno;
     bool crashOnError;

     bool isTmpFile;
     string tmpFileName;
public:

#define MAX_OFF_T ((1 << ((sizeof(off_t) * 8) - 2)))

     class PositionGuard {
	  File *guardedFile;
	  loff_t position;
     public:
	  PositionGuard(File *f) {
	       guardedFile = f;
	       if (guardedFile->Tell(position) != SUCCESS) {
		    assert(0);
	       }
	  }
	  ~PositionGuard() {
	       if (guardedFile->Seek(position) != SUCCESS) {
		    assert(0);
	       }
	  }
     };

     static File Stdout;
     static File Stdin;
     static File Stderr;
     
     static const int    READ_WRITE_FLAGS;
     static const int    READ_FLAGS;
     static const int    WRITE_FLAGS;
     static const mode_t DEFAULT_MODE;

     typedef enum {ERROR = 0, END_OF_FILE, SUCCESS} FileStatus;

     static int tmpFileCount;

     /** 
      * Default constructor.  It does nothing.
      */
     File(bool crashOnError = true);

     /**
      * Constructor to make a File out of a FILE*.  In theory, you should still
      * be able to manupulate the the FILE * directly.  However, this has not
      * been thoroughly tested.
      */
     File(FILE *f, bool crashOnError = false);

     /**
      * Constructor to make a File out of a file descriptor.  The string mode
      * is a mode string that will be passed to fdopen() to create a FILE * for
      * internal use.
      */
     File(int fd, char *mode, bool crashOnError = true);

     /**
      * Open the file with the given path and the given flags.  The flags are
      * the same as those passed to open(2).
      *
      * FIXME: Need to be able to specify permissions as well.
      */
     FileStatus Open(const char *name,   const int flags = READ_WRITE_FLAGS, const mode_t mode = DEFAULT_MODE);
     FileStatus Open(const string &name, const int flags = READ_WRITE_FLAGS, const mode_t mode = DEFAULT_MODE);

     FileStatus OpenTmpFile(char *tmpDir = "", int flags = READ_WRITE_FLAGS | O_TRUNC);
     FileStatus DestroyTmpFile();
     bool IsTmpFile() {return isTmpFile;}
     /**
      * These associate a new file with this object.
      */
     FileStatus SetFile(FILE *f);
     FileStatus SetFile(int fd, char* mode);
     
     /**
      * Write n bytes to the file from the buffer, buf.  If written != NULL,
      * written = number of bytes written.
      */
     FileStatus Write(const void *buf, const size_t n, size_t *written = NULL);
     
     /**
      * Write count items of size size from buffer buf.  If written != NULL,
      * written = number of bytes written.
      */
     virtual FileStatus WriteN(const void *buf, const size_t size, const size_t count, size_t *written = NULL);
     
     /**
      * Write the contents of f to this file
      */
     FileStatus WriteFileContents(File &f, loff_t bytes = ULONG_MAX, loff_t *written = NULL);

     /**
      * Read n bytes into buf.  If read != NULL, *read = bytes actually read.
      * Returns END_OF_FILE if end of file occurs.
      */
     FileStatus Read(void *buf, const size_t n, size_t * read = NULL);
     
     /**
      * Read count items of size size into buf.  If read != NULL, *read = number of items read.
      * Returns END_OF_FILE if end of file occurs.
      */
     virtual FileStatus ReadN(void *buf, const size_t size, const size_t count, size_t *read = NULL);

     /**
      * Read one line into buf.
      * Returns END_OF_FILE if end of file occurs.
      */
     FileStatus Read(string &buf, const char delimiter = '\n');

     /**
      * Read a character.
      * Returns END_OF_FILE if end of file occurs.
      */
     FileStatus Read(char &b);
     
     /**
      * Write a string.
      */
     FileStatus Write(const string &buf);
     
     /**
      * Write a charactor
      */
     FileStatus Write(const char c);

     /**
      * Print formatted text, ala printf(4).  It returns the number of charactors written.
      */
     int Print(const char *fmt, ...) __attribute__ ((format (printf, 2, 3)));

     /**
      * Print formated text, ala printf(4), but use a va_list instead, ala vprintf(4).
      */
     int Print(const char *fmt, va_list ap);

     /**
      * Reposition the file pointer offset bytes from the location specified by
      * whence.  Possible values for whence are SEEK_CUR, SEEK_SET, SEEK_END.
      * For more info see lseek(2).
      */
     FileStatus Seek(const int64 offset, const int whence = SEEK_SET);
     /* This is a bit of hack.  It assumes that off_t is signed.
	Fortunately,if this number is low it will only make things slower, not
	incorrect.  It is only used in Seek() */

     FileStatus Tell(loff_t &l);
     /**
      * Relocate the file pointer to the beginning of the file.
      */
     virtual FileStatus Rewind();

     /**
      * Is the file open?
      */
     bool IsOpen() {return fFile != NULL;}

     /**
      * Close the file.
      */
     FileStatus Close();

     /**
      * Get the FILE * for this File.
      */
     FILE *GetFILE();
     
     /**
      * Get the file descriptor for this File.
      */
     int GetFD();

     /**
      * Crash out because of an error, if crashOnError == true.  Make a
      * subclass and override this if you want different behavior on errors.
      * The default behavior is to print out the most recent error string and
      * then call abort(2).
      */
     virtual void Crash();

     /**
      * The destructor calls Close() and discards the result.  A more
      * conservative implementation would make it an error for the destructor
      * to be called without closing the file.  
      *
      * Note that the destructor does not close the file.
      */
     virtual ~File() {}

     bool operator== (const File f) const {return f.fDes == fDes;}
};
#endif
