#ifndef SPARSE64BIT_MAP_INCLUDED
#define SPARSE64BIT_MAP_INCLUDED
#include"Foundation/Debug.h"
#include"Util/BigVector.h"

extern unsigned long long allocated;
extern unsigned long long freed;
extern unsigned long long allocatedMem;
extern unsigned long long freedMem;

class Sparse64BitMapStats: public StatsContainer {
 public:
     
     Counter bytesAllocated;
     Counter levelOneAllocations;
     Counter levelTwoAllocations;
     Counter itemsInserted;
     Counter reads;
     Counter writes;

     BigVectorStats cacheStats;

     Sparse64BitMapStats(char*name="Sparse64BitMap"): cacheStats("cache") {
	  ADD_STAT_SHORT(bytesAllocated);
	  ADD_STAT_SHORT(levelOneAllocations);
	  ADD_STAT_SHORT(levelTwoAllocations);
	  ADD_STAT_SHORT(itemsInserted);
	  ADD_STAT_SHORT(reads);
	  ADD_STAT_SHORT(writes);
	  
	  AddStatContainer(&cacheStats);
     }
     
     void Reset() {
	  bytesAllocated = 0;
	  levelOneAllocations = 0;
	  levelTwoAllocations = 0;
	  itemsInserted = 0;
	  reads = 0;
	  writes = 0;
	  cacheStats.Reset();
     }

     void AddPrefix(const string &prefix) {
	  StatsContainer::AddPrefix(prefix);
	  cacheStats.AddPrefix(prefix + "cache/");
     }
     
};

template <class TYPE, int MemoryPageSizeBits>
class Sparse64BitMap {
 public:
     uint64 Spaces;
     uint64 Worlds;
     uint64 Pages;
     uint64 PageSize;
     uint64 SpaceBits;
     uint64 WorldBits;
     uint64 PageBits;
     uint64 PageSizeBits;

     uint64 SpaceShift;
     uint64 WorldShift;
     uint64 PageShift;
     uint64 SpaceMask;
     uint64 WorldMask;
     uint64 PageMask;
     uint64 PageSizeMask;
     Counter Accessges;
     Counter BytesRead;
     Counter BytesWritten;
     //VersionedArray<void *> SpaceArray;
     int64 ***SpaceArray;
     bool theIgnoreWrites;

     uint64 AllocatedSize;

     uint LastTransactionSize;
     uint64 LastTransactionAddress;
     
     Sparse64BitMapStats _stats;
     Sparse64BitMapStats * stats;
     
#define MEMORY_PAGE_SIZE (1 << MemoryPageSizeBits)

     struct MemoryPage {
	  uint64 address;
	  uint64 size;

	  TYPE items[MEMORY_PAGE_SIZE];

	  MemoryPage() {
	       size = MEMORY_PAGE_SIZE;
	  }

	  TYPE & operator[](uint64 index) {
	       return items[index];
	  }
	  
	  static size_t SizeOnDisk() {
	       return sizeof(uint64) + sizeof(TYPE) * MEMORY_PAGE_SIZE;
	  }
	  
#define TRY(a) do {					\
     File::FileStatus r = a;			       	\
     if (r != File::SUCCESS) {			        \
	  return r;					\
     }							\
} while(0)
	  File::FileStatus Read(File &f) {
	       TRY(f.Read(&address, sizeof(uint64)));
	       TRY(f.Read(items, sizeof(TYPE) * MEMORY_PAGE_SIZE));
	       return File::SUCCESS;
	  }

	  File::FileStatus Write(File &f) {
	       TRY(f.Write(&address, sizeof(uint64)));
	       TRY(f.Write(items, sizeof(TYPE) * MEMORY_PAGE_SIZE));
	       return File::SUCCESS;
	  }
#undef TRY
     };

     BigVector<MemoryPage> pageCache;

 public:
     Sparse64BitMap(uint64 aBlockSize = 1, uint64 cacheSize = 32, char *tmpDir = "") : 
	  pageCache(0, aBlockSize, cacheSize, tmpDir) {
	  Spaces = Worlds = Pages = PageSize = 0;
	  SpaceArray = NULL;
	  theIgnoreWrites = false;
	  AllocatedSize = 0;
	  SetStats(NULL);
     } 

     ~Sparse64BitMap() {
	  Clear();
     }

     void SetStats(Sparse64BitMapStats *s) {
	  if (s == NULL) {
	       stats = &_stats;
	       pageCache.SetStats(NULL);
	  }  else {
	       stats = s;
	       pageCache.SetStats(&stats->cacheStats);
	  }
     }

     bool Clear() {
	  uint64 Index;

	  AllocatedSize = 0;
	  if (SpaceArray) {
	       for (Index = 0; Index < Spaces; Index++) {
		    if (SpaceArray[Index]) {
			 uint64 Index2;
			 for (Index2 = 0; Index2 < Worlds; Index2++) {
			      if (SpaceArray[Index][Index2]) {
				   freed++;
				delete [] SpaceArray[Index][Index2];
			      }
			 }
			 freed++;
			 delete [] SpaceArray[Index];
		    }
	       }
	       freed++;
	       delete [] SpaceArray;
	       SpaceArray = NULL;
	  }
	  pageCache.Clear();
	  Spaces = Worlds = Pages = PageSize = 0;
	  SpaceBits = WorldBits = PageBits = PageSizeBits = 0;

	  BytesRead = 0;
	  BytesWritten = 0;
	  
	  _stats.Reset();

	  return (true);
     }

     bool Initialize(uint64 aSpaceBits,
		     uint64 aWorldBits,
		     uint64 aPageBits) {
	  uint64 Index;
	  uint64 aPageSizeBits = MemoryPageSizeBits;
	  Clear();

	  if ((aSpaceBits + aWorldBits + aPageBits + aPageSizeBits) != 64) {
	       printf("Error " format_dec_uint64 ":" format_dec_uint64 ":"
		      format_dec_uint64 ":" format_dec_uint64 " != 64 bits\n",
		      aSpaceBits, aWorldBits, aPageBits, aPageSizeBits);
	       return (false);
	  }
	  SpaceBits = aSpaceBits;
	  WorldBits = aWorldBits;
	  PageBits = aPageBits;
	  PageSizeBits = aPageSizeBits;

	  Pages = (1 << (PageBits));
	  Worlds = (1 << (WorldBits));
	  Spaces = (1 << (SpaceBits));

	  PageSize = (1 << (PageSizeBits));

	  SpaceShift = WorldBits + PageBits + PageSizeBits;
	  SpaceMask = (1 << SpaceBits) - 1;
	  PageShift = PageSizeBits;
	  PageMask = (1 << PageBits) - 1;
	  WorldShift = PageBits + PageSizeBits;
	  WorldMask = (1 << WorldBits) - 1;
	  PageSizeMask = (1 << PageSizeBits) - 1;

	  //printf("SpaceBits:    " format_dec_int64 " SpaceShift " format_dec_int64 " SpaceMask: " format_hex_uint64 "\n", SpaceBits, SpaceShift, SpaceMask);
	  //printf("WorldBits:    " format_dec_int64 " WorldShift " format_dec_int64 " WorldMask: " format_hex_uint64 "\n", WorldBits, WorldShift, WorldMask);
	  //printf("PageBits:     " format_dec_int64 "  PageShift " format_dec_int64 "  PageMask: " format_hex_uint64 "\n", PageBits, PageShift, PageMask);
	  //printf("PageSizeBits: " format_dec_int64 "            PageSizeMask: " format_hex_uint64 "\n", PageSizeBits, PageSizeMask);

	  allocated++;
	  SpaceArray = new int64 **[Spaces];
	  stats->bytesAllocated += Spaces * sizeof(int64**);
	  Index = 0;
	  for (Index = 0; Index < Spaces; Index++)
	       SpaceArray[Index] = NULL;

	  return (true);
     }

     bool IgnoreWrites() {
	  theIgnoreWrites = true;
	  return (true);
     }
     bool DoNotIgnoreWrites() {
	  theIgnoreWrites = false;
	  return (true);
     }

     MemoryPage *GetPage(uint64 sa) {
	  uint64 tIndex = (sa >> SpaceShift) & SpaceMask;
	  uint64 tIndex2 = (sa >> WorldShift) & WorldMask;
	  uint64 tIndex3 = (sa >> PageShift) & PageMask;
	  uint64 Index2, Index3;

	  if (theIgnoreWrites)
	       return (NULL);	// speculative mode -- bad address

	  if (!SpaceArray[tIndex]) {
	       allocated++;
	       SpaceArray[tIndex] = new int64 *[Worlds];
	       stats->bytesAllocated += Worlds * sizeof(int64*);
	       //	       memset(SpaceArray[tIndex], 0x00, Worlds * sizeof(int64*));
	       for (Index2 = 0; Index2 < Worlds; Index2++)
		    SpaceArray[tIndex][Index2] = NULL;
	  }

	  if (!SpaceArray[tIndex][tIndex2]) {
	       allocated++;

	       SpaceArray[tIndex][tIndex2] = new int64 [Pages];
	       stats->bytesAllocated += Pages * sizeof(int64);
	       stats->levelOneAllocations++;
	       //memset(SpaceArray[tIndex][tIndex2], 0xff, Pages * sizeof(int64));
	       for (Index3 = 0; Index3 < Pages; Index3++)
		    SpaceArray[tIndex][tIndex2][Index3] = -1;
	  }
	
	  if (SpaceArray[tIndex][tIndex2][tIndex3] == -1) {
	       MemoryPage &m =  pageCache.ElementAt(pageCache.Size());
	       stats->levelTwoAllocations++;
	       m.address = sa;
	       SpaceArray[tIndex][tIndex2][tIndex3] = pageCache.Size() - 1;
	       AllocatedSize = AllocatedSize + PageSize;
	  }

	  return &pageCache[SpaceArray[tIndex][tIndex2][tIndex3]];
     }

     uint GetLastTransactionSize() {
	  return (LastTransactionSize);
     }
     uint64 GetLastTransactionAddress() {
	  return (LastTransactionAddress);
     }

     TYPE Get(uint64 sa) {
	  LastTransactionSize = 1;
	  LastTransactionAddress = sa;
	  MemoryPage *Page = GetPage(sa);
	  if (!Page)
	       return (0);

	  stats->reads++;
	  return ((*Page)[sa & PageSizeMask]);
     }

     TYPE * GetPtr(uint64 sa) {
	  LastTransactionSize = 1;
	  LastTransactionAddress = sa;
	  MemoryPage *Page = GetPage(sa);
	  if (!Page)
	       return (0);

	  stats->reads++;
	  return &((*Page)[sa & PageSizeMask]);
     }

     bool Set(uint64 sa, const TYPE & Data) {
	  LastTransactionSize = 1;
	  LastTransactionAddress = sa;
	  if (theIgnoreWrites)
	       return (true);
	  MemoryPage *Page = GetPage(sa);
	  if (!Page)
	       return (false);

	  (*Page)[sa & PageSizeMask] = Data;
	  stats->writes++;
	  return (true);
     }
     

     BigVector<MemoryPage> *GetBigVector() {
	  return & pageCache;
     }
};

#ifdef TEST_64BIT_SPARSE_MAP

#include <stdlib.h>

struct Test {

     uint v;
     
     Test(uint n = 0) {
	  v = n;
     }

     static size_t SizeOnDisk() {
	  return sizeof(uint);
     }

     File::FileStatus Read(File &f) {
	  return f.Read(&v, sizeof(uint));
     }

     File::FileStatus Write(File &f) {
	  return f.Write(&v, sizeof(uint));
     }
};

int main (int argc, char *argv[0]) 
{
     Sparse64BitMap<Test, 4> v(16, 256);
     uint64 i;
     int k;
     int s;
     v.Initialize(20,20,20);
#define COUNT 100000
     
     for (k = 0; k < 5; k++) {
	  File::Stderr.Print("Filling\n");
	  for (i = 0; i < COUNT; i++) {
	       v.Set(i + k*COUNT, Test(i + k*COUNT));
	  }

	  File::Stderr.Print("Accessing\n");	  
	  for ( i = 0; i < COUNT; i++) {
	       uint j=(int) (((k*COUNT) + 0.0)*rand()/(RAND_MAX+1.0));
	       Test a = v.Get(j);
	       assert(a.v == j);
	  }
     }
     
     v.DumpStats(File::Stdout);
     return (s%2) + 1;
}

#endif
#endif
