#ifndef BASIC_BLOCK_INCLUDED
#define BASIC_BLOCK_INCLUDED

#include <glib.h>
#include <stdio.h>

#include "set.h"
#include "types.h"
#include "ptrSet.h"

#define MAX_EXITS 128
#define MAX_ENTRANCES 128

typedef struct _BASIC_BLOCK *PBasicBlock;

typedef enum {BRANCHING_BEHAVIOR_NONE, /* straight-line */
	      BRANCHING_BEHAVIOR_SIMPLE, /* Taken, not-taken variety */
	      BRANCHING_BEHAVIOR_MULTIWAY, /* many possible targets */
	      BRANCHING_BEHAVIOR_PROCEDURE_CALL, 
	      BRANCHING_BEHAVIOR_MAX} BranchingBehavior;

#include "cfg.h"
#include "function.h"
#include "inst.h"
#include "dominance.h"
#include "lightCycleStats.h"

#define BB_OWNER_PROGRAM(b) ((b)->owner->owner)
#define BB_OWNER_FUNC(b) ((b)->owner)


#define EDGE_BLOCK(a) (a->target)
#define EDGE_FUNCTION(a) (assert(0), NULL) //(a == NULL ? NULL :(PFunction)(a)->target)

#define forEachExit(b, link, var) for((link) = ((b)->exits), (var) = (((b)->exits) == NULL ? NULL : ((PCfgEdge)(b->exits)->data)->dst) ; (link) != NULL; (link) = g_list_next(link), (var) = ((link) == NULL ? NULL : (PCfgEdge)(link)->data)->dst)

#define forEachEntrance(b, link, var) for((link) = ((b)->entrances), (var) = (((b)->entrances) == NULL ? NULL : ((PCfgEdge)(b->entrances)->data)->src) ; (link) != NULL; (link) = g_list_next(link), (var) = ((link) == NULL ? NULL : ((PCfgEdge)(link)->data)->src))


typedef struct _LOOP {
     PBasicBlock head;
     PPtrSet nodes;
     PPtrSet exitEdges;
     PPtrSet exitBlocks;
} Loop, *PLoop;

typedef struct _BASIC_BLOCK {
     GList *exitLabels;

     /* CFG data */
     GList *exits; 
     GList *entrances;

     GList *indirectExits;
     GList *potentialFunctionArgs;

     /* Control distribution data */
     GList *controlEntrances;  	/* these are basic blocks that provide enable messages to this block. */
     GList *controlExits; 	/* This block provides enable messages to these blocks. */
     
     /* control dependence data */
     PSet controlDependentNodeSet;
     PSet controlDependentClosureNodeSet;
     GList *controlDependentNodes;
     GList *controlDependentOnNodes;

     /* Strict control dependence data */
     /* This is my own invention.  A node B strictly control dependent on A if
	there is a path from A that includes B and another that does not
	include B (this i normal control dependence) The additional condition
	is neither of the paths may include A (except at the beginning and at B if B=A).  The
	goal is to make the code following a loop be control dependent on the
	loop condition. */  
     PSet strictControlDependentNodeSet;

     PC entrancePC;
     PFunction owner;

     bool_t handWaveNumber;

     char *label;
     GList *insts;
     
     BranchingBehavior branchingBehavior;

     bool_t isArtificial;  /* This is true for the head and tail blocks we need
			      to add for ssa analysis */
     
     bool_t isSimpleBranch; /* The taken, not-taken variety */

     bool_t isDead;  /* can this basic block be removed */

     // These are for dfg.c
     unsigned long int visited;
     unsigned long int startedHere;

     int id;
     int localID;
     //int isEntryBlock;
     int length;

     int needsJumpPorts;
     int needsReturnPorts;
     PInst jumpInst;

     GArray writerSummary;
     GArray readerSummary;

     DominanceData dominance[DOMINANCE_DIRECTION_MAX];
     
     PInst controlSource;

     GList *minEntrySet;

     GList *noLocalSuccessorInsts;     // Possible control tails.
     GList *noLocalPredecessorInsts;     // Possible control heads.
     GList *noLocalProducerArgs;

     PSet assignedVariables;
     PSet consumedVariables;

     bool_t isFunctionCallBlock;

     // Loop data
     PLoop loop; // It is not possible for a basic block to the head of more
		 // than one loop.  Code in cfg.c ensures this.     
     bool_t isLoopTop;

     bool_t isLoopExitBlock;
     PBasicBlock exitBlockFor;

     // Epoch Data
     bool_t isEpochHead;
     PPtrSet epoch;
     bool_t isInEpoch;
     PBasicBlock containingEpoch;
     PPtrSet epochEdge;
     bool_t hasEpochExitsInstalled;
     PPtrSet registersUsed;

     // Reachablity data
     // First index is forward or backward, second index is with our without back edges
     PPtrSet reachableNodes[2][2];
     
     // These are just for the ssa analysis
     int ssaWork;
     int ssaHasAlready;

     StatsBasicBlock stats;

     int extraInstructions;

     GList *rippleChunks;

} BasicBlock;



PBasicBlock basicBlockNew(char *label, int blockID);
void basicBlockFree(PBasicBlock block);
void basicBlockPrint(FILE *f, PBasicBlock block, bool_t verbose);
void basicBlockPrintAnnotatedMap(PBasicBlock block, FILE * file);

void basicBlockSetPC(PBasicBlock block, PC pc) ;
void basicBlockSetJumpFlags(PBasicBlock block);
void basicBlockSetDoesJump(PBasicBlock block);
void basicBlockAddIndirectExit(PBasicBlock from, PBasicBlock to);
void basicBlockSetNeedsPorts(PBasicBlock block);

void basicBlockAppendInst(PBasicBlock block, PInst inst);
void basicBlockPrependInst(PBasicBlock block, PInst inst);
void basicBlockInsertInstAfterInst(PBasicBlock block, PInst after, PInst inst);
void basicBlockInsertInstBeforeInst(PBasicBlock block, PInst before, PInst inst);
void basicBlockRemoveInst(PBasicBlock block, PInst inst);

void basicBlockFindUsedVariables(PBasicBlock block);

void basicBlockClearVisitedFlags (GList *blocks) ;
GList * basicBlockLastInstLink(PBasicBlock b) ;

PBasicBlock basicBlockGetTakenExit(PBasicBlock block);
PBasicBlock basicBlockGetNotTakenExit(PBasicBlock block);

PInst basicBlockLastInst(PBasicBlock b);
char* basicBlockName_DOT(PBasicBlock block, char *buf);
void basicBlockPrint_DOT(FILE *f, PBasicBlock block, bool_t includeDfg);

void basicBlockFindPotentialFunctionParams(PBasicBlock block);
void basicBlockAddCallPorts(PBasicBlock block);
void basicBlockAddReturnPort(PBasicBlock block);

void basicBlockAddJumpPorts(PBasicBlock block);
void basicBlockAddReturnPort(PBasicBlock block);

void basicBlockComputeDataFlowEntryPoints(PBasicBlock block);
void basicBlockComputeDataFlowExitPoints(PBasicBlock block);

void basicBlockAddControlDependenceEdges(PBasicBlock block);
void basicBlockSetControlSource(PBasicBlock block);

void basicBlockAddControlDistributionEdge(PBasicBlock b, PBasicBlock currentBlock);
void basicBlockSanity(PBasicBlock block);

void basicBlockComputeNoLocalSuccessorInsts(PBasicBlock block) ;
void basicBlockComputeNoLocalPredecessorInsts(PBasicBlock block);

PCfgEdge basicBlockGetNotTakenExitEdge(PBasicBlock block);
PCfgEdge basicBlockGetTakenExitEdge(PBasicBlock block);
PInst basicBlockExitInst(PBasicBlock b) ;
void  basicBlockMerge(PBasicBlock top, PBasicBlock bottom);

void basicBlockBuildRippleChunks(PBasicBlock b);


#endif
