#include "basicBlock.h"
#include <assert.h>
#include "inst.h"
#include <glib.h>
#include "list.h"
#include "dot.h"
#include "cfg.h"

CfgOptions cfgOptions;

void cfgPrint_DOT(FILE *f, PFunction func, bool_t includeDfg) 
{
     GList *l;
     PBasicBlock block;
     char graphName[1024];
     sprintf(graphName, "%s: CFG", func->name);

     dotOpenGraph(f, graphName);
     
     forEach(func->basicBlocks, l, block) {
	  basicBlockPrint_DOT(f, block, includeDfg);
     }

     dotCloseGraph(f);
}

void cfgPrintExport(PFunction func, FILE *f) 
{

     GList *l1, *l2;
     PBasicBlock block;
     PCfgEdge exit;

     forEach(func->basicBlocks, l1, block) {
	  int first = 1;

	  fprintf(f, "%d:", block->id);
	  forEach(block->exits, l2, exit) {
	       if (!first) {
		    fprintf(f, ",");
	       } else {
		    first = 0;
	       }
	       fprintf(f, "%d", exit->dst->id);
	  }
	  fprintf(f, "\n");
     }
}


PCfgEdge cfgEdgeNew(PBasicBlock src, PBasicBlock dst, CfgEdgeType type)
{
     PCfgEdge n = (PCfgEdge)calloc(sizeof(CfgEdge), 1);
     
     n->type = type;
     n->src = src;
     n->dst = dst;
     return n;
}

void cfgEdgeFree(PCfgEdge e)
{
     free(e);
}

PCfgEdge cfgAddEdge(PBasicBlock from, PBasicBlock to, CfgEdgeType type) 
{
     PCfgEdge e;
     assert(to && from);
     e = cfgEdgeNew(from, to, type);
     from->exits = g_list_append(from->exits, e);
     if (to) {
	  to->entrances = g_list_append(to->entrances, e);
     }
     return e;
}

void cfgRemoveBlock(PBasicBlock block)
{

     PCfgEdge e;
     GList *l;
     GList *copy = NULL;

     copy = g_list_copy(block->exits);
     forEach(copy, l, e) {
	  cfgRemoveEdge(e);
     }
     g_list_free(copy);

     copy = g_list_copy(block->entrances);
     forEach(copy, l, e) {
	  cfgRemoveEdge(e);
     }
     g_list_free(copy);
}

void cfgRemoveEdge(PCfgEdge e) 
{
     e->dst->entrances = g_list_remove(e->dst->entrances, e);
     e->src->exits = g_list_remove(e->src->exits, e);
     e->dst = NULL;
     e->src = NULL;
}

void cfgEdgeMoveSrc(PCfgEdge e, PBasicBlock newSrc) 
{
     assert(newSrc);
     e->src->exits = g_list_remove(e->src->exits, e);
     newSrc->exits = g_list_append(newSrc->exits, e);
     e->src = newSrc;
}

void cfgEdgeMoveDst(PCfgEdge e, PBasicBlock newDst) 
{
     assert(newDst);
     e->dst->entrances = g_list_remove(e->dst->entrances, e);
     newDst->entrances = g_list_append(newDst->entrances, e);
     e->dst = newDst;
}

void cfgBuild(PFunction func)
{
     GList *i;
     GList *childless = NULL;
     PBasicBlock block, first;
     int op;

     forEachLink(func->basicBlocks, i) {
	  PInst exitInst;

	  block = (PBasicBlock)i->data;
	  exitInst = basicBlockExitInst(block);
	  if (exitInst != NULL) {

	       MD_SET_OPCODE(op, exitInst->binInst);
	       
	       if (instIsCondBranch(exitInst)) {
		    cfgAddEdge(block, programFindBasicBlockByPC(FUNC_OWNER_PROGRAM(func), exitInst->targetPC), CFG_EDGE_TAKEN);
		    if (g_list_next(i)) {
			 cfgAddEdge(block, (PBasicBlock)g_list_next(i)->data, CFG_EDGE_NOT_TAKEN);
		    } else {
			 assert(0);
		    }
	       } else if (op == BR) {
		    cfgAddEdge(block, programFindBasicBlockByPC(FUNC_OWNER_PROGRAM(func), exitInst->targetPC), CFG_EDGE_NORMAL);
	       } else {
		    if (g_list_next(i)) {
			 cfgAddEdge(block, (PBasicBlock)g_list_next(i)->data, CFG_EDGE_NORMAL);
		    } 
	       }
	  }
     }

     
     // This is a hack.  We need all entry points.  This should be it.
     first = functionFindBasicBlockByLocalID(func, 2);
     
     cfgAddEdge(func->enter, first, CFG_EDGE_NORMAL);
     cfgAddEdge(func->enter, func->exit, CFG_EDGE_NORMAL);
     
     forEach(func->basicBlocks, i, block) {
	  if (g_list_empty(block->exits)) {
	       childless = g_list_append(childless, block);
	  }
     }
     
     forEach(childless, i, block) {
	  if (block != func->exit) {
	       cfgAddEdge(block, func->exit, CFG_EDGE_NORMAL);
	  }
     }
     
}

bool_t cfgStrictlyDominates(PBasicBlock a, PBasicBlock b, DominanceDirection direction)
{
     GList *l;
     PBasicBlock block;
     
     // This is an ineffecient pile of crap.
     
     forEach(a->dominance[direction].dominates, l, block) {
	  bool_t r;

	  if (block == b) {
	       return TRUE;
	  }

	  r = cfgDominates(block, b, direction);
	  if  (r == TRUE) {
	       return TRUE;
	  }
     }

     return FALSE;
     
}

bool_t cfgDominates(PBasicBlock a, PBasicBlock b, DominanceDirection direction)
{
     if (a == b) {
	  return TRUE;
     }
     
     return cfgStrictlyDominates(a, b, direction);

}

void cfgReachableTearDown(PFunction func)
{
     GList *i;
     int m,n;
     PBasicBlock block;

     for (n = 0; n < 2; n++) {
	  for (m = 0; m < 2; m++) {
	       forEach(func->basicBlocks, i, block) {
		    if (block->reachableNodes[n][m]) {
			 ptrSetFree(block->reachableNodes[n][m]);
		    }
		    block->reachableNodes[n][m] = NULL;
	       }
	  }
     }
}

void cfgLabelBackEdges(PFunction func)
{
     GList *i, *l;
     PBasicBlock block;
     PCfgEdge e;

     forEach(func->basicBlocks, i, block) {    
	  forEach(block->entrances, l, e) {
	       if (cfgDominates(e->dst, e->src, DOMINANCE_DIRECTION_PRE)) {
		    e->isBackEdge = TRUE;
	       } else {
		    e->isBackEdge = FALSE;
	       }
	  }
     }
}

void cfgComputeRechableSets(PFunction func, int dir, bool_t allowBackEdges)
{

     GList *workList = NULL;
     PBasicBlock block;
     GList *i;

     forEach(func->basicBlocks, i, block) {
	  block->reachableNodes[dir][allowBackEdges] = ptrSetNew();
	  ptrSetAdd(block->reachableNodes[dir][allowBackEdges], block);
     }
	       
     forEach(func->basicBlocks, i, block) {
	  workList = g_list_enqueue(workList, block);	  
     }
 
     while (workList != NULL) {
	  PCfgEdge e;
	  GList *outs;
	  block = (PBasicBlock)workList->data;
	  workList = g_list_dequeue(workList);
	  

	  // BACKWARD uses the exits instead of the entrances because we
	  // collect data in the opposite direction.
	  outs = dir == REACHABLE_BACKWARD ? block->exits : block->entrances;

	  forEach(outs, i, e) {
	       int sizeBefore;
	       int sizeAfter;

	       PBasicBlock next = dir == REACHABLE_BACKWARD ? e->dst : e->src;
	       
	       if (!allowBackEdges) {
		    if(e->isBackEdge) {
			 continue;
		    }
	       }
	       
	       sizeBefore = ptrSetSize(next->reachableNodes[dir][allowBackEdges]);
	       ptrSetUnion(next->reachableNodes[dir][allowBackEdges], 
			   next->reachableNodes[dir][allowBackEdges], 
			   block->reachableNodes[dir][allowBackEdges]);
	       sizeAfter = ptrSetSize(next->reachableNodes[dir][allowBackEdges]);

	       if (sizeBefore != sizeAfter) {
		    workList = g_list_enqueue(workList, next);
	       }
	  }
     }
     
}


void cfgRemoveEmptyBlocks(PFunction func)
{
     GList *l;
     PBasicBlock block;
     GList *c;

     c = g_list_copy(func->basicBlocks);

     forEach(c, l, block) {
	  if (block == func->exit || block == func->enter) {
	       continue;
	  }

	  if(g_list_length(block->insts) == 0) {
	       PCfgEdge e;
	       PCfgEdge theExit;
	       GList *c2;
	       GList *i;
	       assert(g_list_length(block->exits) == 1);
	       
	       theExit = (PCfgEdge)block->exits->data;

	       c2 = g_list_copy(block->entrances);
	       forEach(c2, i, e) {
		    cfgEdgeMoveDst(e, theExit->dst);
	       }
	       g_list_free(c2);

	       cfgRemoveEdge(theExit);
	       cfgRemoveBlock(block);
	       functionRemoveBasicBlock(func, block);
	       basicBlockFree(block);
	  }
     }

     g_list_free(c);
}

PLoop loopNew(PBasicBlock head)
{
     PLoop new;

     nice_calloc(new, sizeof(Loop));
     
     new->nodes = ptrSetNew();
     new->exitEdges = ptrSetNew();
     new->exitBlocks = ptrSetNew();
     new->head = head;
     return new;
}

void loopFree(PLoop loop) 
{
     ptrSetFree(loop->nodes);
     ptrSetFree(loop->exitEdges);
     free(loop);
}

typedef struct _LOOP_SIZE {
     PCfgEdge edge;
     int backEdgeSrcCount;
     PLoop loop;
} LoopSize, *PLoopSize;

static PLoopSize newLoopSize(PCfgEdge e, PLoop loop)
{
     PLoopSize n;

     nice_calloc(n, sizeof(LoopSize));
     n->edge = e;
     n->loop = loop;

     return n;
}


gint cmpLoopNest(gconstpointer aa, gconstpointer bb) 
{
     PLoopSize a = (PLoopSize)aa;
     PLoopSize b = (PLoopSize)bb;
     
     if (a->backEdgeSrcCount < b->backEdgeSrcCount) {
	  return -1;
     }  else if ( a->backEdgeSrcCount > b->backEdgeSrcCount) {
	  return 1;
     } else {
	  return 0;
     }
}

static void cfgComputeReachableBackwardWithExclusion(PPtrSet reachable, PBasicBlock start, PBasicBlock exclude)
{
     GList *work = NULL;
     PBasicBlock block;
     PCfgEdge e;
     

     ptrSetEmpty(reachable);

     // POSSIBLE FATALITY.  I added this code it might break things and the
     // current 6/14/2002 regressions won't catch it.
     if (start == exclude) {
	  ptrSetAdd(reachable, start);
	  return;
     }
     work = g_list_push(work, start);

     while (work) {
	  GList *l;
	  block = (PBasicBlock)g_list_top(work);
	  work = g_list_pop(work);

	  ptrSetAdd(reachable, block);

	  forEach(block->entrances, l, e) {
	       if (e->src != exclude && !ptrSetElementOf(reachable, e->src)) {
		    work = g_list_push(work, e->src);
	       }
	  }
     }
}

static void loopUpdateNodes(PLoop loop, PCfgEdge edge)
{
     
     cfgComputeReachableBackwardWithExclusion(loop->nodes, edge->src, edge->dst);
     ptrSetAdd(loop->nodes, edge->dst);
     
#if (0)
     ptrSetIntersect(loop->nodes, edge->src->reachableNodes[REACHABLE_BACKWARD][FALSE], 
		     edge->dst->dominance[DOMINANCE_DIRECTION_PRE].dominatedNodes);


     fprintf(stderr, "EDGE: %d -> %d\nREACHABLE(%d): ", edge->src->id, edge->dst->id, edge->src->id);

     forEachPtrElement(edge->src->reachableNodes[REACHABLE_BACKWARD][FALSE], i, b) {
	  fprintf(stderr, " %d\t", b->id);
     }

     fprintf(stderr, "\nDOMINATED(%d): ", edge->dst->id);

     forEachPtrElement(edge->dst->dominance[DOMINANCE_DIRECTION_PRE].dominatedNodes, i, b) {
	  fprintf(stderr, " %d\t", b->id);
     }

     fprintf(stderr, "\nRESULT: ");
     
     forEachPtrElement(loop->nodes, i, b) {
	  fprintf(stderr, " %d\t", b->id);
     }

     fprintf(stderr, "\n\n");
#endif     
}

static PLoop loopConstruct(PCfgEdge edge)
{
     PLoop loop = loopNew(edge->dst);
     GList *l, *j;
     PBasicBlock loopMember;

     loopUpdateNodes(loop, edge);

     // We need to collect all the exits from the loop.  All the nodes
     // that branch out of the loop go into exits.
     forEachPtrElement(loop->nodes, l, loopMember) {
	  PCfgEdge e;
	  forEach(loopMember->exits, j, e) {
	       if (!ptrSetElementOf(loop->nodes, e->dst)) {
		    ptrSetAdd(loop->exitEdges, e);
		    e->isLoopExitEdge = TRUE;
	       }
	  }
     }

     return loop;
}

gint cmpLoopNesting(gconstpointer a, gconstpointer b)
{
     PBasicBlock aa = (PBasicBlock)a;
     PBasicBlock bb = (PBasicBlock)b;
     bool_t AContainsB = ptrSetContains(aa->loop->nodes, bb->loop->nodes);
     bool_t BContainsA = ptrSetContains(bb->loop->nodes, aa->loop->nodes);

     
     if ((BContainsA && AContainsB) ||  // Shouldn't happen because we give each loop its own head.
	 (!BContainsA && !AContainsB) ) {  // shouldn't happen because we are only supposed to be comparing nested loops
	  assert(0); 
	  return 0;
     } else if (BContainsA) {
	  return -1;
     } else if (AContainsB) {
	  return 1;
     } else {
	  assert(0);
	  return 0;
     }
}

void cfgAddExitBlocks(PFunction func)
{ 
     GList *i;
     PtrSetIndex l;
     PBasicBlock block;
     PPtrSet visited = ptrSetNew();

     forEach(func->basicBlocks, i, block) {
	  if (block->loop) {
	       PCfgEdge e;
	       forEachPtrElement(block->loop->exitEdges, l, e) {
		    PBasicBlock new = basicBlockNew(NULL, -1);
		    functionAddBasicBlock(func, new);
		    cfgAddEdge(new, e->dst, CFG_EDGE_NORMAL);
		    cfgEdgeMoveDst(e, new);
		    ptrSetAdd(block->loop->exitBlocks, new);
		    new->exitBlockFor = block;
		    new->isLoopExitBlock = TRUE;
	       }
	  }
     }

     // We have add the exit blocks, but there is a catch.  If there is a
     // branch out of two loops, there will be two exit blocks there.  This is
     // fine, except they may not be in the correct order.  So we have to find
     // all the chains of exit blocks and rearrange them so they are in the
     // correct order.
     forEach(func->basicBlocks, i, block) {
	  GList *controllingLoops = NULL;
	  GList *exitBlocks = NULL;
	  PBasicBlock exitBlock;
	  GList *k;

	  // We are only interested in the heads of the chains of exit nodes
	  if (g_list_length(block->entrances) == 1 && ((PCfgEdge)block->entrances->data)->src->isLoopExitBlock) {
	       continue;
	  }

	  if (!block->exitBlockFor) {
	       continue;
	  }

	  exitBlock = block;
	  while (exitBlock != NULL) {
	       ptrSetAdd(visited, exitBlock);
	       assert(exitBlock->exitBlockFor);

	       controllingLoops = g_list_append(controllingLoops, exitBlock->exitBlockFor);
	       exitBlocks = g_list_append(exitBlocks, exitBlock);

	       //	       fprintf(stderr, "removing %d from %d\n", exitBlock->id, exitBlock->exitBlockFor->id);
	       ptrSetRemove(exitBlock->exitBlockFor->loop->exitBlocks, exitBlock);
	       exitBlock->exitBlockFor = NULL;

	       if (g_list_length(exitBlock->exits) == 1 && ((PCfgEdge)exitBlock->exits->data)->dst->isLoopExitBlock) {
		    exitBlock = ((PCfgEdge)exitBlock->exits->data)->dst;
	       } else {
		    exitBlock = NULL;
	       }
	  }
	       
	  controllingLoops = g_list_sort(controllingLoops, cmpLoopNesting);
	       
	  k = controllingLoops;
	  forEach(exitBlocks, l, exitBlock) {
	       assert(k);
	       //	       fprintf(stderr, "Adding %d to %d\n", exitBlock->id, ((PBasicBlock)k->data)->id);
	       ptrSetAdd(((PBasicBlock)k->data)->loop->exitBlocks, exitBlock);
	       exitBlock->exitBlockFor = (PBasicBlock)k->data;
	       k = k->next;
	  }
	  assert(k == NULL);
	       
	  g_list_free(controllingLoops);
     }

     ptrSetFree(visited);
     
}

void cfgTearDownLoops(PFunction func)
{
     GList *i;
     PBasicBlock block;

     forEach(func->basicBlocks, i, block) {
	  if (block->loop) {
	       loopFree(block->loop);
	       block->loop = NULL;
	  }
     }
}

void cfgUpdateLoopBodies(PFunction func)
{
     GList *l, *i;
     PBasicBlock block;
     PCfgEdge e;

     forEach(func->basicBlocks, i, block) {
	  forEach(block->entrances, l, e) {
	       if (e->isBackEdge) {
		    loopUpdateNodes(block->loop, e);
	       }
	  }
     }
}

void cfgFindLoops(PFunction func)
{
     GList *l, *i;
     PBasicBlock block;
     PCfgEdge e;

     forEach(func->basicBlocks, i, block) {
	  forEach(block->entrances, l, e) {
	       if (e->isBackEdge) {
		    assert(block->loop == NULL);
		    block->loop = loopConstruct(e);
	       }
	  }
     }
}


void cfgExpandLoops(PFunction func)
{
     PBasicBlock block;
     GList *i, *l, *blocksCopy;
     PCfgEdge e;

     blocksCopy = g_list_copy(func->basicBlocks);
     forEach(blocksCopy, i, block) {    
	  PPtrSet backEdgeSrcs = ptrSetNew();
	  GList  *backEdgeEntrances = NULL;
	  GList *copy;
	  PBasicBlock currentChild;
	  GList *k;
	  PCfgEdge firstAdded = NULL;
	  PLoopSize loopSize;
	  PLoop loop;


	  // We look at each backedge coming into this block.  For each edge
	  // compute the loop that it covers
	  //	  fprintf(stderr, "LOOPS @ %d:\n", block->id);
	  forEach(block->entrances, l, e) {
	       if (cfgDominates(e->dst, e->src, DOMINANCE_DIRECTION_PRE)) {

		    ptrSetAdd(backEdgeSrcs, e->src);
		    e->isBackEdge = TRUE;
		    
		    //loopBody = cfgFindLoop(e);
		    loop = loopConstruct(e);

		    backEdgeEntrances = g_list_append(backEdgeEntrances, newLoopSize(e, loop));

/* 		    fprintf(stderr, "\t"); */
/* 		    forEachPtrElement(loop->nodes, m, blarg) { */
/* 			 fprintf(stderr, " %d", blarg->id); */
/* 		    }		     */
/* 		    fprintf(stderr, "\n"); */
	       }
	  }

	  // Once we have all the loops we have to order them in nested order.  We
	  // do this by computing the number of the back edge sources that each
	  // loop contains.  The outermost loop will contain them all, the next
	  // in, all but one, and so on.

/* 	  fprintf(stderr, "\t"); */
/* 	  forEachPtrElement(backEdgeSrcs, m, blarg) { */
/* 	       fprintf(stderr, " %d", blarg->id); */
/* 	  }		     */
/* 	  fprintf(stderr, "\n"); */
	  
	  forEach(backEdgeEntrances, l, loopSize) {
	       PPtrSet t = ptrSetNew();
	       ptrSetIntersect(t, loopSize->loop->nodes, backEdgeSrcs);
	       loopSize->backEdgeSrcCount = ptrSetSize(t);
	       ptrSetFree(t);
	  }

	  backEdgeEntrances = g_list_sort(backEdgeEntrances, cmpLoopNest);
	  
	  // Now we add new basic blocks for the headers of each loop, and fix
	  // up the cfg.  We hoist the back edges as we go, when we are done
	  // with all them, we hoist all the inputs to the original block to
	  // the input of the outermost loop head.
	  currentChild = block;
	  forEach_r(backEdgeEntrances, l, loopSize) {
	       PBasicBlock new = basicBlockNew(NULL, -1);
	       PCfgEdge e;
	       //	       fprintf(stderr, "Adding loop head @ %d for src %d\n", block->id, loop->edge->src->id);
	       new->isLoopTop = TRUE;
	       
	       functionAddBasicBlock(func, new);

	       e = cfgAddEdge(new, currentChild, CFG_EDGE_NORMAL);

	       if (firstAdded == NULL) {
		    firstAdded = e;
	       }
	       
	       cfgEdgeMoveDst(loopSize->edge, new);

	       free(loopSize);
	       currentChild = new;
	  }
	  
	  copy = g_list_copy(block->entrances);
	  forEach(copy, k, e) {
	       if (e != firstAdded) {
		    cfgEdgeMoveDst(e, currentChild);
	       }
	  }
	  g_list_free(copy);
	  

	  g_list_free(backEdgeEntrances);
	  ptrSetFree(backEdgeSrcs);
     }

     g_list_free(blocksCopy);

}

