#include "ssa.h"
#include "basicBlock.h"
#include "inst.h"
#include "function.h"
#include "set.h"
#include "list.h"
#include "assert.h"

static PCfgEdge findDependenceExitEdge_HELP(PSet visited, PBasicBlock src, PBasicBlock cur)
{
     GList *l;
     PBasicBlock predecessor;
     PCfgEdge entrance;
     PCfgEdge ret = NULL;

     forEach(cur->entrances, l, entrance) {
	  predecessor = entrance->src;

	  if (!setElementOf(visited, predecessor->localID)) {
	       setAdd(visited, predecessor->localID);	       
	       if (predecessor == src) {
		    ret = entrance; 
	       } else {
		    PCfgEdge e = findDependenceExitEdge_HELP(visited, src, predecessor);

		    if (e != NULL) {
			 ret = e;
		    }
	       }
	  } else {
	       setAdd(visited, predecessor->localID);	       
	  }

	  
	  if (ret != NULL) {
	       return ret;
	  }
     }
     
     return NULL;
}

/* 
   Figure 14 from Cryton.  Building the set first and then building the list
   gaurentees uniqueness in the list
*/
static void buildControlDependenceSets(PFunction func)
{
     GList *l1;
     PBasicBlock x, y;
     int x1;

     forEach(func->basicBlocks, l1, x) {
	  x->controlDependentNodeSet = setNew(func->lengthBasicBlocks + 1);
     } 
     
     forEach(func->basicBlocks, l1, y) {
	  forEachElement(y->dominance[DOMINANCE_DIRECTION_POST].dominanceFrontier, x1) {
	       x = FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, x1);
	       setAdd(x->controlDependentNodeSet, y->localID);
	  }
     }

     forEach(func->basicBlocks, l1, y) {
	  forEachElement(y->controlDependentNodeSet, x1) {
	       x = FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, x1);
	       y->controlDependentNodes = g_list_append(y->controlDependentNodes, x);
	       x->controlDependentOnNodes = g_list_append(x->controlDependentOnNodes, y);
	  }
     }
     
}


// This function is currently used but is a prototypical dataflow analysis.
static void buildControlDependenceClosureSets(PFunction func)
{
     GList *workList = NULL;
     PBasicBlock block;
     GList *i;
     PSet visited = setNew(func->lengthBasicBlocks + 1);

     forEach(func->basicBlocks, i, block) {
	  block->controlDependentClosureNodeSet = setNew(func->lengthBasicBlocks + 1);
	  setCopy(block->controlDependentClosureNodeSet, block->controlDependentNodeSet);
     }
	       
     forEach(func->basicBlocks, i, block) {
	  //	  fprintf(stderr, "Added %d\n", block->id);
	  workList = g_list_enqueue(workList, block);	  
     }
 
     while (workList != NULL) {
	  PBasicBlock parent;
	  
	  block = (PBasicBlock)workList->data;
	  workList = g_list_dequeue(workList);
	  setAdd(visited, block->localID);
	  //	  fprintf(stderr, "Processing %d\n", block->id);
	  
	  forEach(block->controlDependentOnNodes, i, parent) {
	       bool_t changed;
	       //	       fprintf(stderr, "Checking %d\n", parent->id);
	       changed = setCombineUnion(parent->controlDependentClosureNodeSet, 
					 block->controlDependentClosureNodeSet); 

	       if (changed || !setElementOf(visited, parent->localID)) {
		    //		    fprintf(stderr, "Added %d\n", parent->id);
		    workList = g_list_enqueue(workList, parent);
	       }
	  }

	  //fprintf(stderr, "Done\n");
     }
     
     setFree(visited);
}

static PCfgEdge findDependenceExitEdge(PBasicBlock src, PBasicBlock dst)
{
     /* figure whether the control flow from src to dst flows through the taken
	or not taken exit from src to dst.
     */

     PSet visited = setNew(src->owner->lengthBasicBlocks + 1);
     PCfgEdge r = findDependenceExitEdge_HELP(visited, src, dst); 
     setFree(visited);
     assert(r != NULL);
     return r;
}

static void partitionControlDependences(PFunction func) 
{
     GList *i, *j;
     PBasicBlock block, dependent;
     PCfgEdge srcEdge;
     PCfgEdge e;

     forEach(func->basicBlocks, i, block) {
	  forEach(block->exits, j, e) {
	       e->controlDependentNodes = setNew(func->lengthBasicBlocks + 1);
	  }
     }

     forEach(func->basicBlocks, i, block) {

	  forEach(block->controlDependentNodes, j, dependent) {
	       srcEdge = findDependenceExitEdge(block, dependent);
	       setAdd(srcEdge->controlDependentNodes, dependent->localID);
	  }
     }
}

static void partitionControlDependenceClosure(PFunction func) 
{
     GList *i, *j;
     PBasicBlock block, dependent;
     PCfgEdge srcEdge;
     PCfgEdge e;
     int v;

     forEach(func->basicBlocks, i, block) {	  
	  forEach(block->exits, j, e) {
	       e->controlDependentClosureNodeSet = setNew(func->lengthBasicBlocks + 1);
	  }

	  forEachElement(block->controlDependentClosureNodeSet, v) {
	       dependent = FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, v);
	       srcEdge = findDependenceExitEdge(block, dependent);
	       setAdd(srcEdge->controlDependentClosureNodeSet, dependent->localID);
	  }
     }
}

bool_t controlDependenceDependsOn(PBasicBlock b1, PBasicBlock b2)
{
     return setElementOf(b2->controlDependentClosureNodeSet, b1->localID);
}

void controlDependencesCompute(PFunction func)
{
     buildControlDependenceSets(func);
     buildControlDependenceClosureSets(func);
     partitionControlDependenceClosure(func);
     partitionControlDependences(func);
}

bool_t strictControlDependenceDependsOn(PBasicBlock b1, PBasicBlock b2)
{
     return setElementOf(b2->strictControlDependentNodeSet, b1->localID);
}


void strictControlDependenceSearch(PBasicBlock src, PBasicBlock b, PSet s)
{
     GList *l;
     PCfgEdge e;

     //     fprintf(stderr, "Traversing %d\n", b->id);
     
     setAdd(s, b->localID);

     forEach(b->exits, l, e) {
	  if (src == e->dst) {
	       setAdd(s, e->dst->localID);
	  } else if (!setElementOf(s, e->dst->localID)) {
	       strictControlDependenceSearch(src, e->dst, s);
	  }
     }
}

void strictControlDependenceCompute(PFunction func)
{
     GList *l;
     PBasicBlock block;

     forEach(func->basicBlocks, l, block) {
	  block->strictControlDependentNodeSet = setNew(func->lengthBasicBlocks + 1);	  
	  if (g_list_length(block->exits) > 1) {
	       GList *sets = NULL, *i;
	       PSet s;
	       PSet intersection, _union; // Damn keywords
	       PCfgEdge e;


	       forEach(block->exits, i, e) {
		    s = setNew(func->lengthBasicBlocks + 1);
		    strictControlDependenceSearch(block, e->dst, s);
		    sets = g_list_append(sets, s);
	       }
	       
	       intersection = setNew(func->lengthBasicBlocks + 1);
	       _union = setNew(func->lengthBasicBlocks + 1);
	       setCopy(intersection, s);  // Note that s is left over from the
					  // loop above.

	       forEach(sets, i, s) {
		    setIntersection(intersection, intersection, s);
		    setUnion(_union, _union, s);
	       }

	       setSubstract(block->strictControlDependentNodeSet, _union, intersection);
	       
	       setFree(_union);
	       setFree(intersection);
	       
	       forEach(sets, i, s) {
		    setFree(s);
	       }
	       
	       g_list_free(sets);
	  }
     }
}
