#include "basicBlock.h"
#include <assert.h>
#include "inst.h"
#include <glib.h>
#include "list.h"
#include "machine.h"
#include "cfg.h"
#include "util.h"


void dfgAddEdge(PArg from, PArg to) 
{
     to->initializedOnce = 1;
     assert(from->valid &&  to->valid);
     from->consumers = g_list_append(from->consumers, to);

     if (to) {
	  to->producers = g_list_append(to->producers, from);
     }
}

void dfgSmartRemoveEdge(PInst src, OutputArgType srcType, PInst dst, InputArgType dstType)
{
     PArg srcArg;
     PArg dstArg;

     srcArg = instGetDstByType(src, srcType);
     dstArg = instGetSrcByType(dst, dstType);

     srcArg->consumers = g_list_remove(srcArg->consumers, dstArg);
     dstArg->producers = g_list_remove(dstArg->producers, srcArg);
}

void dfgRemoveInstEdges(PInst inst)
{
     ArgIndex i;
     PArg src, dst;
     GList *l;

     forEachSrc(inst, i, src) {
	  PArg producer;
	  GList *copy = g_list_copy(src->producers);
	  forEach(copy, l, producer) {
	       dfgRemoveEdge(producer, src);
	  }
	  g_list_free(copy);
     }

     forEachDst(inst, i, dst) {
	  PArg consumer;
	  GList *copy = g_list_copy(dst->consumers);

	  forEach(copy, l, consumer) {
	       dfgRemoveEdge(dst, consumer);
	  }
	  g_list_free(copy);
     }

}

void dfgSmartAddEdge (PInst src, OutputArgType srcType, PInst dst, InputArgType dstType)
{
     
     PArg srcArg;
     PArg dstArg;
     int v;

     srcArg = instGetDstByType(src, srcType);
     dstArg = instGetSrcByType(dst, dstType);
     
     if (srcArg == NULL &&
	 dstArg == NULL) {
	  v = getUniqueVariable();
     } else {

	  if (srcArg != NULL) {
	       v = srcArg->reg;
	  } else if (dstArg != NULL) {
	       v = dstArg->reg;
	  }
	  
     }

     if (srcArg == NULL) {
	  srcArg = instAddDst(src, v, srcType);
     }

     if (dstArg == NULL) {
	  dstArg = instAddSrc(dst, v, dstType);
     }

     dfgAddEdge(srcArg, dstArg);
}

void dfgRemoveEdge(PArg src, PArg dst)
{
     src->consumers = g_list_remove(src->consumers, dst);
     dst->producers = g_list_remove(dst->producers, src);     
}

static unsigned long int currentEpoch = 0;

static void findRAWHazard(GList* fromInstLink, PArg src,  PBasicBlock b, int doFirst) 
{
     GList* i;
     ArgIndex j;
     PArg dst;
     PInst inst;
     PCfgEdge entrance;
     
     if (b->visited == currentEpoch && b->startedHere != currentEpoch) {
	  return;
     }
     
     //     fprintf(stderr, "Examining %s\n", b->label);


     if (b->visited) {
	  b->startedHere = currentEpoch - 1;
     }

     b->visited = currentEpoch;

     // doFirst is so we can skip the instruction we are called with.
     // this is needed for only the first call to searchUpFromSource
     // while searching for a particular instruction.  It prevents an
     // automatic self-dependence.
     if (!doFirst) {
	  i =  g_list_previous(fromInstLink);
     } else {
	  i = fromInstLink;
     }

     for (; i != NULL; i = g_list_previous(i)) {
    
	  inst = (PInst)i->data;

	  //    	  fprintf(stderr, "checking Inst: %d ", inst->id);    
/*    	  md_print_insn(inst->binInst, 0, stderr);   */
    	  // fprintf(stderr, "\n");   
	  
	  forEachDst(inst, j, dst) {
	       //	       if (ARG_IS_VALUE(dst)) {
	       if (dst->reg == src->reg) {			 
/*  		    fprintf(stderr, "FOUND DEP (%d)\n", dst->reg);  */
		    dfgAddEdge(dst, src);
		    return;
	       }
		    //	       }
	  }
     }


     if (g_list_empty(b->entrances)) {
	  src->notAlwaysInitialized = 1;
	  return;
     } else { 
	  forEach(b->entrances, i, entrance) {
	       assert(entrance != NULL);
	       findRAWHazard(basicBlockLastInstLink(entrance->src),
			     src,
			     entrance->src,
			     1);
	  }
     }
}

void dfgUpdateSource(PArg src) 
{ 
     PInst inst = src->owner;
     PBasicBlock block = inst->owner;
     GList *instLink;

     instLink = g_list_find(block->insts, inst);
     block->startedHere = ++currentEpoch;
     findRAWHazard(instLink, src, block, 0);
}

static void dfgBuildEdges(GList *blocks, PBasicBlock b, GList *instLink)
{
     PInst inst;
     PArg src;
     ArgIndex i;

     inst = (PInst)instLink->data;

     forEachSrc(inst, i, src) {
	  if (src->reg != 31 && src->reg != -1) {
	       b->startedHere = ++currentEpoch;
	       findRAWHazard(instLink, src, b, 0);
	  }
     }
}

void dfgBuild(PFunction func)
{
     GList *l;
     GList *i;
     PBasicBlock currentBlock;

     forEach(func->basicBlocks, l, currentBlock) {
	  currentBlock = (PBasicBlock)l->data;    
	  
	  // 	  fprintf(stderr, "STARTING  %s\n", currentBlock->label);		     
	  forEachLink(currentBlock->insts, i) {
 	       PInst inst = (PInst)i->data;

	       //	       fprintf(stderr, "STARTING INST %d\n", inst->id);		     
	       dfgBuildEdges(func->basicBlocks, currentBlock, i);
	  }
     }
}


void dfgDismantal(PFunction func)
{
     GList *l1, *l2;
     PBasicBlock block;
     PInst inst;
     PArg arg;
     ArgIndex i;

     forEach(func->basicBlocks, l1, block) {
	  forEach(block->insts, l2, inst) {
	       forEachSrc(inst, i, arg) {
		    argClearDfgPtrs(arg);
	       }
	       forEachDst(inst, i, arg) {
		    argClearDfgPtrs(arg);
	       }
	  }
     }
}

void dfgPrintExport(PFunction func, FILE *f) 
{
     GList *l1, *l2, *l3;
     ArgIndex i;
     PBasicBlock block;
     PInst inst;
     PArg src, dst, producer;
     //     PArg argA, argC;
     bool_t first;
	       
     forEach(func->basicBlocks, l1, block) {
	  forEach(block->insts, l2, inst) {

	       fprintf(f, "%lu:", inst->id);
	       
	       first = TRUE;
	       forEachDst(inst, i, dst) {
		    if (ARG_VALID(dst) && dst->reg != -1) {
			 if (!first) fprintf(f, ",");
			 fprintf(f, "%d", dst->reg);
			 first = FALSE; 
		    } else {
			 continue;
		    }
	       }

	       if (inst->type == OP_sink) {
		    PArg input = instGetSrcByType(inst, INPUT_A);
		    if (!first) fprintf(f, ",");
		    fprintf(f, "%d", input->reg);
	       }
	       fprintf(f, ":");

	       forEachSrc(inst, i, src) {
		    first = TRUE;
		    if (ARG_VALID(src) && src->reg != -1) {
			 fprintf(f, "%d", src->reg);
			 first = FALSE; 
		    } else {
			 fprintf(f, ":");
			 continue;
		    }

		    forEach(src->producers, l3, producer) {
			 if (!first) {
			      fprintf(f, ",");
			 } else {
			      first = 0;
			 }
			 fprintf(f, "%lu", producer->owner->id);
		    }
		    fprintf(f, ":");
	       }


	       if (inst->type == OP_src) {
		    assert(0);
		    PArg output = instGetDstByType(inst, OUTPUT_C);
		    fprintf(f, "%d:", output->reg);
	       }
	       
	       fprintf(f, "\n");
	  }
     }
}

void dfgCollectStats(PFunction func, PDfgStats stats) 
{
     GList *l1, *l2, *l3;
     ArgIndex i;
     PBasicBlock block;
     PInst inst;
     PArg dst, consumer;
     PtrSetIndex j;
     PBasicBlock epochMember;
     
     forEach(func->basicBlocks, l1, block) {
	  forEach(block->insts, l2, inst) {
	       int switchOutDegree = 0;
	       int controlOutDegree = 0;
	       int dataOutDegree = 0;
	       int totalOutDegree;

	       if (instIsCondBranch(inst)) {
		    stats->branchCount++;
	       } 
	       stats->instCount++;
	       forEachDst(inst, i, dst) {
		    forEach(dst->consumers, l3, consumer) {
			 switch(consumer->inputType) {
/* 			 case INPUT_SWITCH: */
/* 			      switchOutDegree++; */
/* 			      stats->switchEdgeCount++; */
/* 			      break; */
/* 			 case INPUT_CONTROL: */
/* 			      controlOutDegree++; */
/* 			      stats->controlEdgeCount++; */
/* 			      break; */
			 case INPUT_A:
			 case INPUT_B:
			      //			 case INPUT_OTHER:
			      dataOutDegree++;
			      stats->dataEdgeCount++;
			 default:
			      break;
			 }

		    }
	       }
	       
	       stats->maxControlOutDegree = MAX(stats->maxControlOutDegree,
						controlOutDegree);
	       stats->maxDataOutDegree = MAX(stats->maxDataOutDegree,
						dataOutDegree);
	       stats->maxSwitchOutDegree = MAX(stats->maxSwitchOutDegree,
						switchOutDegree);

	       totalOutDegree = controlOutDegree + dataOutDegree + switchOutDegree;
	       
	       stats->maxTotalOutDegree = MAX(stats->maxTotalOutDegree,
					      totalOutDegree);
	       
	       stats->instructionDistribution[inst->type]++;
	       
	       if (inst->latencyType & (LATENCY_BRANCHES|
					LATENCY_EASY_EPOCH_ADVANCES|
					LATENCY_EASY_RHO_FUNCS|
					LATENCY_ZAPPED_EPOCH_MANAGEMENT)) {
		    stats->zeroLatencyInstructionCount++;
	       } else {
		    if (inst->type != OP_TYPE_ALPHA_OP) {
			 stats->overheadInstructions++;
		    }
	       }
	       
	  }

	  stats->extraInstructionCount += block->extraInstructions;
		    
	  if (block->isEpochHead) {
	       int instCount = 0, blockCount = 0;
	       forEachPtrElement(block->epoch, j, epochMember) {
		    blockCount++;
		    instCount += g_list_length(epochMember->insts);
	       }
	       stats->epochBasicBlockSizeHistogram[MIN(EPOCH_BASIC_BLOCK_SIZE_HISTOGRAM_BUCKETS - 1, blockCount)]++;
	       stats->epochInstSizeHistogram[MIN(EPOCH_OP_SIZE_HISTOGRAM_BUCKETS - 1, instCount)]++;
	       
	  }
	  
	  
     }
}


void dfgPrintStats(FILE *f, PDfgStats stats)
{
     int s;
     InstType type;
     int total;
     //     int i;
     fprintf(f, "Inst Count:                 %4d\n", stats->instCount);
     fprintf(f, "Branch Count:               %4d(%4.2f%%)\n", 
	     stats->branchCount, 
	     PERCENT(stats->branchCount, stats->instCount));

     fprintf(f, "\n");
     s = stats->dataEdgeCount + stats->controlEdgeCount + stats->switchEdgeCount;
     fprintf(f, "total edges:                %4d\n", s);
     fprintf(f, "data edges:                 %4d(%4.2f%%)\n", 
	     stats->dataEdgeCount, 
	     PERCENT(stats->dataEdgeCount, s));
     fprintf(f, "control edges:              %4d(%4.2f%%)\n", 
	     stats->controlEdgeCount, 
	     PERCENT(stats->controlEdgeCount, s));
     fprintf(f, "switch edges:               %4d(%4.2f%%)\n", 
	     stats->switchEdgeCount, 
	     PERCENT(stats->switchEdgeCount, s));
     fprintf(f, "\n");

     fprintf(f, "Average outdegree:          %4.2f\n", RATIO(s, stats->instCount));
     fprintf(f, "Average data outdegree:     %4.2f\n", RATIO(stats->dataEdgeCount, stats->instCount));
     fprintf(f, "Average control degree:     %4.2f\n", RATIO(stats->controlEdgeCount, stats->branchCount));
     fprintf(f, "Average switch outdegree:   %4.2f\n", RATIO(stats->switchEdgeCount, stats->branchCount));

     fprintf(f, "\n");
     fprintf(f, "Max data out degree:        %4d\n", stats->maxDataOutDegree);
     fprintf(f, "Max control out degree:     %4d\n", stats->maxControlOutDegree);
     fprintf(f, "Max switch out degree:      %4d\n", stats->maxSwitchOutDegree);
     fprintf(f, "Max total out degree:       %4d\n", stats->maxTotalOutDegree);

     fprintf(f, "\n");
     
     fprintf(f, "Zero latency instructions:  %4d (%5.2f)\n", stats->zeroLatencyInstructionCount, PERCENT(stats->zeroLatencyInstructionCount, stats->instCount));
     fprintf(f, "Removed noops:              %4d (%5.2f)\n", stats->extraInstructionCount, PERCENT(stats->extraInstructionCount, stats->instCount + stats->extraInstructionCount));
     

     total = 0;
     for(type = 0; type < OP_TYPE_MAX; type++) {
	  total += stats->instructionDistribution[type];
     }

     fprintf(f, "Static instruction distribution\n");
     for(type = 0; type < OP_TYPE_MAX; type++) {
	  fprintf(f, "%10s  :  %9d(%5.1f)\n", 
		  OP_TYPE_NAME[type], 
		  stats->instructionDistribution[type], 
		  PERCENT(stats->instructionDistribution[type], total));
     }
     fprintf(f, "%10s  :  %9d\n", "Total", total);
     fprintf(f, "%10s  :  %9d\n", "Overhead", stats->overheadInstructions);
     fprintf(f, "\n");

#if (0)     
     fprintf(f, "Epoch Basic Block Size Distribution:\n");
     for(i = 0; i < EPOCH_BASIC_BLOCK_SIZE_HISTOGRAM_BUCKETS; i++) {
	  fprintf(f, "%4d : %d\n", i, stats->epochBasicBlockSizeHistogram[i]);
     }
     fprintf(f, "\n");

#endif
#if(1)
     int i;
     fprintf(f, "Epoch Inst Size Distribution:\n");
     for(i = 0; i < EPOCH_OP_SIZE_HISTOGRAM_BUCKETS; i++) {
	  fprintf(f, "%4d : %d\n", i, stats->epochInstSizeHistogram[i]);
     }
     fprintf(f, "\n");
#endif

}
