#include "ssa.h"
#include "list.h"
#include "set.h"
#include "util.h"
#include <assert.h>
#include "set.h"
#include "dot.h"
#include "buildInst.h"


/* The next few static variables and functions and
   dominanceComputeImmediateDominators is a direct implementation of the algorithm by
   Tarjan from ACM Transactions on Programming Languages and Systems, Vol 1, No
   1, July 1979. 

   It has been extended slightly to allow for computation of forward and
   reverse domination information.  It is black magic.
*/

static int *parent, *ancestor, *vertex;
static int *label, *semi;
static PSet *pred, *bucket;
static int *dom;
static guint32 u, v;
static int n, r;

static void DFS(PBasicBlock b, DominanceDirection dir)
{
     PCfgEdge exit;
     GList *childList;
     GList *l;
     int v = b->localID;

     n = n + 1;
     semi[v] = n;
     vertex[n] = label[v] = v;
     ancestor[v] = 0;
 
     childList = (dir == DOMINANCE_DIRECTION_PRE) ? b->exits : b->entrances;

     forEach(childList, l, exit) {
	  PBasicBlock e;
	  int w;
	  e = (dir == DOMINANCE_DIRECTION_PRE) ? exit->dst : exit->src;
	  w = e->localID;
	  assert(w != 0);
	  if (semi[w] == 0) {
	       parent[w] = v;
	       DFS(e, dir);
	  }
	  setAdd(pred[w], v);
     }
}

static void COMPRESS(int v)
{
     if (ancestor[ancestor[v]] != 0) {
	  COMPRESS(ancestor[v]);
	  if (semi[label[ancestor[v]]] < semi[label[v]]) {
	       label[v] = label[ancestor[v]];
	  }
	  ancestor[v] = ancestor[ancestor[v]];
     }
}

static int EVAL(int v)
{
     if (ancestor[v] == 0) {
	  return v;
     } else {
	  COMPRESS(v);
	  return label[v];
     }
}

static void LINK(int v, int w)
{
     ancestor[w] = v;
}


static int* DOMINATORS (PFunction func, DominanceDirection dir)
{
     int w;
     int i;
     PBasicBlock root;

     n = func->lengthBasicBlocks;
     nice_calloc(parent,   (n + 1) * sizeof(int));
     nice_calloc(ancestor, (n + 1) * sizeof(int));
     nice_calloc(vertex,   (n + 1) * sizeof(int));
     nice_calloc(label,    (n + 1) * sizeof(int));
     nice_calloc(semi,     (n + 1) * sizeof(int));
     nice_calloc(dom,      (n + 1) * sizeof(int));

     nice_calloc(pred,     (n + 1) * sizeof(PSet));
     nice_calloc(bucket,   (n + 1) * sizeof(PSet));

     n = func->lengthBasicBlocks;

     if (dir == DOMINANCE_DIRECTION_PRE) {
	  root = FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, 1);
     } else {
	  root = FUNC_LAST_BASIC_BLOCK(func);
     }
     r = root->localID;

     // Step 1
     for (v = 1; v <= n; v++) {
	  pred[v] = setNew(n + 1);
	  bucket[v] = setNew(n + 1);
	  semi[v] = 0;
     }
     
     n = 0;
     
     DFS(root, dir);
     
     for (i = n; i >= 2; i--) {
	  w = vertex[i];

	  // Step 2
	  forEachElement(pred[w], v) {
	       u = EVAL(v); 
	       if (semi[u] < semi[w]) {
		    semi[w] = semi[u];
	       } 
	  }
	  
	  setAdd(bucket[vertex[semi[w]]], w);
	  
	  LINK(parent[w], w);
	  
	  // Step 3
	  forEachElement(bucket[parent[w]], v) {
	       setRemove(bucket[parent[w]], v);
	       u = EVAL(v);
	       if (semi[u] < semi[v]) {
		    dom[v] = u;
	       } else {
		    dom[v] = parent[w];
	       }
	  }
     }

     // step 4
     for (i = 2; i <= n; i++) {
	  w = vertex[i];
	  if (dom[w] != vertex[semi[w]]) {
	       dom[w] = dom[dom[w]];
	  }
     }

     dom[r] = 0;

     free(parent);
     free(ancestor);
     free(vertex);
     free(label);
     free(semi);

     for (v = 1; v <= n; v++) {
	  setFree(pred[v]);
	  setFree(bucket[v]);
     }

     return dom;
}

static void addDominationEdge(PBasicBlock from, PBasicBlock to, DominanceDirection dir)
{
     to->dominance[dir].dominator = from;
     from->dominance[dir].dominates = g_list_append(from->dominance[dir].dominates, to);
}



/* The next three functions implement figure 10, of "Efficiently Computing
   Static Single Assignment Form and the Control Dependence Graph", Cryton et.al
   March 7, 1991.  The algorithm calls for a bottom of traversal of the
   dominance graph.  We use depth-first post-order. */
static void dominanceComputeDominanceFrontiers_WORK(PBasicBlock x, DominanceDirection dir)
{
     GList *l1, *l2, *childList;
     PBasicBlock y;
     PBasicBlock z;
     PCfgEdge p;
     int y2;
     PFunction func = x->owner;

/*      fprintf(stderr, "Traversing %d\n", x->localID); */

     setEmpty(x->dominance[dir].dominanceFrontier);
     
     childList = (dir == DOMINANCE_DIRECTION_PRE) ? x->exits : x->entrances;

     forEach(childList, l1, p) {
	  y = (dir == DOMINANCE_DIRECTION_PRE) ? p->dst : p->src;	 

	  if (y->dominance[dir].dominator != x) {
	       setAdd(x->dominance[dir].dominanceFrontier, y->localID);
	  }
     }
     
     forEach(x->dominance[dir].dominates, l2, z) {
	  forEachElement(z->dominance[dir].dominanceFrontier, y2) {
	       y = FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, y2);
	       if (y->dominance[dir].dominator != x) {
		    setAdd(x->dominance[dir].dominanceFrontier, y->localID);
	       }
	  }
     }
}

static void dominanceComputeDominanceFrontiers_TRAVERSE(PBasicBlock x, DominanceDirection dir)
{
     GList *l;
     PBasicBlock block;

     forEach(x->dominance[dir].dominates, l, block) {
	  dominanceComputeDominanceFrontiers_TRAVERSE(block, dir);
     }

     dominanceComputeDominanceFrontiers_WORK(x, dir);
     
}

void dominanceComputeDominanceFrontiers(PFunction func, DominanceDirection dir)
{
     GList *l;
     PBasicBlock b;

     forEach(func->basicBlocks, l, b) {
	  b->dominance[dir].dominanceFrontier = setNew(func->lengthBasicBlocks + 1);
     }
     
     if (dir == DOMINANCE_DIRECTION_PRE) {
	  dominanceComputeDominanceFrontiers_TRAVERSE(FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, 1), dir);
     } else {
	  dominanceComputeDominanceFrontiers_TRAVERSE(FUNC_LAST_BASIC_BLOCK(func), dir);
     }
}


void dominancePrintDominatorTree(PFunction func, FILE *f, DominanceDirection dir) 
{
     char srcName[1024], dstName[1024];
     GList *l1, *l2;
     PBasicBlock src;
     PBasicBlock dst;
     char graphName[1024];

     sprintf(graphName, "%s: %s", func->name, dir == DOMINANCE_DIRECTION_PRE ? "Pre dominance" : "Post dominance");
     dotOpenGraph(f, graphName);

     forEach(func->basicBlocks, l1, src) {
	  fprintf(f, "%s;\n",  basicBlockName_DOT(src, srcName));
	  forEach(src->dominance[dir].dominates, l2, dst) {
	       fprintf(f, "%s -> %s [weight=100];\n",  basicBlockName_DOT(dst, dstName), basicBlockName_DOT(src, srcName)); 
	  }
     }

     dotCloseGraph(f);
}

void dominancePrintCDG(PFunction func, FILE *f)
{
     char srcName[1024], dstName[1024];
     GList *l1, *l2;
     PBasicBlock src;
     PBasicBlock dst;
     char graphName[1024];

     sprintf(graphName, "%s: Control dependence", func->name);
     dotOpenGraph(f, graphName);

     forEach(func->basicBlocks, l1, src) {
	  fprintf(f, "%s;\n",  basicBlockName_DOT(src, srcName));
	  forEach(src->controlDependentNodes, l2, dst) {
	       fprintf(f, "%s -> %s [weight=100];\n",  basicBlockName_DOT(src, srcName), basicBlockName_DOT(dst, dstName)); 
	  }
     }
     dotCloseGraph(f);
}

static void dominanceComputeDominanceClosure(PBasicBlock block, DominanceDirection dir)
{
     GList *l;
     PBasicBlock b;

     block->dominance[dir].dominatedNodes = ptrSetNew();

     ptrSetAdd(block->dominance[dir].dominatedNodes, block);
     if (g_list_length(block->dominance[dir].dominates) != 0) {
	  forEach(block->dominance[dir].dominates, l, b) {
	       dominanceComputeDominanceClosure(b, dir);
	       ptrSetUnion(block->dominance[dir].dominatedNodes, block->dominance[dir].dominatedNodes, b->dominance[dir].dominatedNodes); 
	  }
     }

}

void dominanceTearDown(PFunction func)
{
     GList *l1;
     PBasicBlock block;
     int i;

     forEach(func->basicBlocks, l1, block) {
	  for(i = 0; i < DOMINANCE_DIRECTION_MAX; i++) {
	       if (block->dominance[i].dominates) {
		    g_list_free(block->dominance[i].dominates);
		    block->dominance[i].dominates = NULL;
	       } 
	       
	       if (block->dominance[i].dominanceFrontier) {
		    setFree(block->dominance[i].dominanceFrontier);
		    block->dominance[i].dominanceFrontier = NULL;		    
	       }
	       
	       if (block->dominance[i].dominatedNodes) {
		    ptrSetFree(block->dominance[i].dominatedNodes);
		    block->dominance[i].dominatedNodes = NULL;
	       }
	  }
     }
}

void dominanceComputeDominatorTree(PFunction func, DominanceDirection dir)
{
     int *dominators;
     GList *l1;
     PBasicBlock block;

     dominators = DOMINATORS(func, dir);
     
     forEach(func->basicBlocks, l1, block) {
	  PBasicBlock dominator;
	  dominator = FUNC_BASIC_BLOCK_BY_LOCAL_ID(func, dominators[block->localID]);
	  
	  if (dominator != NULL) {
	       addDominationEdge(dominator, block, dir);
	  }
     }
     free(dominators);
     
     if (dir == DOMINANCE_DIRECTION_PRE) {
	  dominanceComputeDominanceClosure(func->enter, dir);
     } else {
	  dominanceComputeDominanceClosure(func->exit, dir);
     }
}
