/*
 * loader.c - program loader routines
 *
 * This file is a part of the SimpleScalar tool suite written by
 * Todd M. Austin as a part of the Multiscalar Research Project.
 *  
 * The tool suite is currently maintained by Doug Burger and Todd M. Austin.
 * 
 * Copyright (C) 1994, 1995, 1996, 1997 by Todd M. Austin
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dburger@cs.wisc.edu
 * US Mail:  1210 W. Dayton Street, Madison, WI 53706
 *
 * $Id: loader.c,v 1.1.1.1 2003/01/09 05:41:09 swanson Exp $
 *
 * $Log: loader.c,v $
 * Revision 1.1.1.1  2003/01/09 05:41:09  swanson
 * Kick it off
 *
 * Revision 1.1.1.1  2000/05/26 15:22:27  taustin
 * SimpleScalar Tool Set
 *
 *
 * Revision 1.3  1999/12/31 18:59:46  taustin
 * quad_t naming conflicts removed
 * cross-endian execution support added (w/ limited syscall support)
 *
 * Revision 1.2  1999/12/13 19:00:08  taustin
 * cross endian execution support added
 *
 * Revision 1.1  1998/08/27 16:54:16  taustin
 * Initial revision
 *
 * Revision 1.1  1998/05/06  01:08:39  calder
 * Initial revision
 *
 * Revision 1.6  1997/04/16  22:09:05  taustin
 * added standalone loader support
 *
 * Revision 1.5  1997/03/11  01:12:39  taustin
 * updated copyright
 * swapping supported disabled until it can be tested further
 *
 * Revision 1.4  1997/01/06  15:59:22  taustin
 * stat_reg calls now do not initialize stat variable values
 * ld_prog_fname variable exported
 *
 * Revision 1.3  1996/12/27  15:51:28  taustin
 * updated comments
 *
 * Revision 1.1  1996/12/05  18:52:32  taustin
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "host.h"
#include "misc.h"
#include "machine.h"
#include "endian.h"
#include "regs.h"
#include "memory.h"
#include "sim.h"
#include "eio.h"
#include "loader.h"

#ifdef BFD_LOADER
#include <bfd.h>
#else /* !BFD_LOADER */
#include "target-alpha/ecoff.h"
#endif /* BFD_LOADER */

/* amount of tail padding added to all loaded text segments */
#define TEXT_TAIL_PADDING 0 /* was: 128 */

/* program text (code) segment base */
md_addr_t ld_text_base = 0;

/* program text (code) size in bytes */
unsigned int ld_text_size = 0;

/* program initialized data segment base */
md_addr_t ld_data_base = 0;

/* top of the data segment */
md_addr_t ld_brk_point = 0;

/* program initialized ".data" and uninitialized ".bss" size in bytes */
unsigned int ld_data_size = 0;

/* program stack segment base (highest address in stack) */
md_addr_t ld_stack_base = 0;

/* program initial stack size */
unsigned int ld_stack_size = 0;

/* lowest address accessed on the stack */
md_addr_t ld_stack_min = -1;

/* program file name */
char *ld_prog_fname = NULL;

/* program entry point (initial PC) */
md_addr_t ld_prog_entry = 0;

/* program environment base address address */
md_addr_t ld_environ_base = 0;

/* target executable endian-ness, non-zero if big endian */
int ld_target_big_endian;

/* register simulator-specific statistics */
void
ld_reg_stats(struct stat_sdb_t *sdb)	/* stats data base */
{
     stat_reg_addr(sdb, "ld_text_base",
		   "program text (code) segment base",
		   &ld_text_base, ld_text_base, "0x%010p");
     stat_reg_uint(sdb, "ld_text_size",
		   "program text (code) size in bytes",
		   &ld_text_size, ld_text_size, NULL);
     stat_reg_addr(sdb, "ld_data_base",
		   "program initialized data segment base",
		   &ld_data_base, ld_data_base, "0x%010p");
     stat_reg_uint(sdb, "ld_data_size",
		   "program init'ed `.data' and uninit'ed `.bss' size in bytes",
		   &ld_data_size, ld_data_size, NULL);
     stat_reg_addr(sdb, "ld_stack_base",
		   "program stack segment base (highest address in stack)",
		   &ld_stack_base, ld_stack_base, "0x%010p");
#if 0 /* FIXME: broken... */
     stat_reg_addr(sdb, "ld_stack_min",
		   "program stack segment lowest address",
		   &ld_stack_min, ld_stack_min, "0x%010p");
#endif
     stat_reg_uint(sdb, "ld_stack_size",
		   "program initial stack size",
		   &ld_stack_size, ld_stack_size, NULL);
     stat_reg_addr(sdb, "ld_prog_entry",
		   "program entry point (initial PC)",
		   &ld_prog_entry, ld_prog_entry, "0x%010p");
     stat_reg_addr(sdb, "ld_environ_base",
		   "program environment base address address",
		   &ld_environ_base, ld_environ_base, "0x%010p");
     stat_reg_int(sdb, "ld_target_big_endian",
		  "target executable endian-ness, non-zero if big endian",
		  &ld_target_big_endian, ld_target_big_endian, NULL);
}


/* load program text and initialized data into simulated virtual memory
   space and initialize program segment range variables */
void
ld_load_prog(char *fname,		/* program to load */
	     int argc, char **argv,	/* simulated program cmd line args */
	     char **envp,		/* simulated program environment */
	     struct regs_t *regs,	/* registers to initialize for load */
	     struct mem_t *mem,		/* memory space to load prog into */
	     int zero_bss_segs)		/* zero uninit data segment? */
{
     int i;
     qword_t temp;
     md_addr_t sp, data_break = 0, null_ptr = 0, argv_addr, envp_addr;
     FILE *fobj;
     long floc;
     struct ecoff_filehdr fhdr;
     struct ecoff_aouthdr ahdr;
     struct ecoff_scnhdr shdr;

     /* record profile file name */
     ld_prog_fname = argv[0];

     /* load the program into memory, try both endians */
     fobj = fopen(argv[0], "r");

     if (!fobj)
	  fatal("cannot open executable `%s'", argv[0]);

     if (fread(&fhdr, sizeof(struct ecoff_filehdr), 1, fobj) < 1)
	  fatal("cannot read header from executable `%s'", argv[0]);

     /* record endian of target */
     if (fhdr.f_magic == MD_SWAPH(ECOFF_ALPHAMAGIC))
	  ld_target_big_endian = FALSE;
     else if (fhdr.f_magic == MD_SWAPH(ECOFF_EB_MAGIC)
	      || fhdr.f_magic == MD_SWAPH(ECOFF_EL_MAGIC)
	      || fhdr.f_magic == MD_SWAPH(ECOFF_EB_OTHER)
	      || fhdr.f_magic == MD_SWAPH(ECOFF_EL_OTHER))
	  fatal("Alpha simulator cannot run PISA binary `%s'", argv[0]);
     else
	  fatal("bad magic number in executable `%s' (not an executable)",
		argv[0]);

     if (fread(&ahdr, sizeof(struct ecoff_aouthdr), 1, fobj) < 1)
	  fatal("cannot read AOUT header from executable `%s'", argv[0]);

     ld_text_base = MD_SWAPQ(ahdr.text_start);
     ld_text_size = MD_SWAPQ(ahdr.tsize);
     ld_prog_entry = MD_SWAPQ(ahdr.entry);
     ld_data_base = MD_SWAPQ(ahdr.data_start);
     ld_data_size = MD_SWAPQ(ahdr.dsize) + MD_SWAPQ(ahdr.bsize);
     regs->regs_R[MD_REG_GP] = MD_SWAPQ(ahdr.gp_value);

     /* compute data segment size from data break point */
     data_break = ld_data_base + ld_data_size;

     /* seek to the beginning of the first section header, the file header comes
	first, followed by the optional header (this is the aouthdr), the size
	of the aouthdr is given in Fdhr.f_opthdr */
     fseek(fobj, sizeof(struct ecoff_filehdr) + MD_SWAPH(fhdr.f_opthdr), 0);

     debug("processing %d sections in `%s'...",
	   MD_SWAPH(fhdr.f_nscns), argv[0]);

     /* loop through the section headers */
     floc = ftell(fobj);
     for (i = 0; i < MD_SWAPH(fhdr.f_nscns); i++) {
	  char *p;

	  if (fseek(fobj, floc, 0) == -1)
	       fatal("could not reset location in executable");
	  if (fread(&shdr, sizeof(struct ecoff_scnhdr), 1, fobj) < 1)
	       fatal("could not read section %d from executable", i);
	  floc = ftell(fobj);

	  switch (MD_SWAPW(shdr.s_flags)) {
	  case ECOFF_STYP_TEXT:
	       p = calloc(MD_SWAPQ(shdr.s_size), sizeof(char));
	       if (!p)
		    fatal("out of virtual memory");

	       if (fseek(fobj, MD_SWAPQ(shdr.s_scnptr), 0) == -1)
		    fatal("could not read `.text' from executable", i);
	       if (fread(p, MD_SWAPQ(shdr.s_size), 1, fobj) < 1)
		    fatal("could not read text section from executable");

	       /* copy program section into simulator target memory */
	       mem_bcopy(mem_access, mem, Write,
			 MD_SWAPQ(shdr.s_vaddr), p, MD_SWAPQ(shdr.s_size));

	       /* release the section buffer */
	       free(p);


	       break;

	  case ECOFF_STYP_INIT:
	  case ECOFF_STYP_FINI:
	       if (MD_SWAPQ(shdr.s_size) > 0) {
		    p = calloc(MD_SWAPQ(shdr.s_size), sizeof(char));
		    if (!p)
			 fatal("out of virtual memory");
		
		    if (fseek(fobj, MD_SWAPQ(shdr.s_scnptr), 0) == -1)
			 fatal("could not read `.text' from executable", i);
		    if (fread(p, MD_SWAPQ(shdr.s_size), 1, fobj) < 1)
			 fatal("could not read text section from executable");
		
		    /* copy program section into simulator target memory */
		    mem_bcopy(mem_access, mem,
			      Write, MD_SWAPQ(shdr.s_vaddr),
			      p, MD_SWAPQ(shdr.s_size));
		
		    /* release the section buffer */
		    free(p);
	       }
	       else
		    warn("section `%s' is empty...", shdr.s_name);
	       break;

	  case ECOFF_STYP_LITA:
	  case ECOFF_STYP_LIT8:
	  case ECOFF_STYP_LIT4:
	  case ECOFF_STYP_XDATA:
	  case ECOFF_STYP_PDATA:
	  case ECOFF_STYP_RCONST:
	       /* fall through */

	  case ECOFF_STYP_RDATA:
	       /* The .rdata section is sometimes placed before the text
		* section instead of being contiguous with the .data section.
		*/
	       /* fall through */
	  case ECOFF_STYP_DATA:
	       /* fall through */
	  case ECOFF_STYP_SDATA:

	       if (MD_SWAPQ(shdr.s_size) > 0) {
		    p = calloc(MD_SWAPQ(shdr.s_size), sizeof(char));
		    if (!p)
			 fatal("out of virtual memory");

		    if (fseek(fobj, MD_SWAPQ(shdr.s_scnptr), 0) == -1)
			 fatal("could not read `.text' from executable", i);
		    if (fread(p, MD_SWAPQ(shdr.s_size), 1, fobj) < 1)
			 fatal("could not read text section from executable");

		    /* copy program section it into simulator target memory */
		    mem_bcopy(mem_access, mem,
			      Write, MD_SWAPQ(shdr.s_vaddr),
			      p, MD_SWAPQ(shdr.s_size));

		    /* release the section buffer */
		    free(p);
	       }
	       else
		    warn("section `%s' is empty...", shdr.s_name);
	       break;

	  case ECOFF_STYP_BSS:
	  case ECOFF_STYP_SBSS:
	       /* no data to read... */
	       break;

	  default:
	       warn("section `%s' ignored...", shdr.s_name);
	  }
     }

     /* done with the executable, close it */
     if (fclose(fobj))
	  fatal("could not close executable `%s'", argv[0]);


     /* perform sanity checks on segment ranges */
     if (!ld_text_base || !ld_text_size)
	  fatal("executable is missing a `.text' section");
     if (!ld_data_base || !ld_data_size)
	  fatal("executable is missing a `.data' section");
     if (!ld_prog_entry)
	  fatal("program entry point not specified");

     /* determine byte/words swapping required to execute on this host */
     sim_swap_bytes = (endian_host_byte_order() != endian_target_byte_order());
     sim_swap_words = (endian_host_word_order() != endian_target_word_order());

     /* set up a local stack pointer, this is where the argv and envp
	data is written into program memory */
     ld_stack_base = ld_text_base - (409600+4096);

     sp = ld_stack_base - MD_MAX_ENVIRON;
     ld_stack_size = ld_stack_base - sp;

     /* initial stack pointer value */
     ld_environ_base = sp;

     /* write [argc] to stack */
     temp = MD_SWAPQ(argc);
     mem_access(mem, Write, sp, &temp, sizeof(qword_t));
     regs->regs_R[MD_REG_A0] = temp;
     sp += sizeof(qword_t);

     /* skip past argv array and NULL */
     argv_addr = sp;
     regs->regs_R[MD_REG_A1] = argv_addr;
     sp = sp + (argc + 1) * sizeof(md_addr_t);

     /* save space for envp array and NULL */
     envp_addr = sp;
     for (i=0; envp[i]; i++)
	  sp += sizeof(md_addr_t);
     sp += sizeof(md_addr_t);

     /* fill in the argv pointer array and data */
     for (i=0; i<argc; i++)
	  {
	       /* write the argv pointer array entry */
	       temp = MD_SWAPQ(sp);
	       mem_access(mem, Write, argv_addr + i*sizeof(md_addr_t),
			  &temp, sizeof(md_addr_t));
	       /* and the data */
	       mem_strcpy(mem_access, mem, Write, sp, argv[i]);
	       sp += strlen(argv[i])+1;
	  }
     /* terminate argv array with a NULL */
     mem_access(mem, Write, argv_addr + i*sizeof(md_addr_t),
		&null_ptr, sizeof(md_addr_t));

     /* write envp pointer array and data to stack */
     for (i = 0; envp[i]; i++)
	  {
	       /* write the envp pointer array entry */
	       temp = MD_SWAPQ(sp);
	       mem_access(mem, Write, envp_addr + i*sizeof(md_addr_t),
			  &temp, sizeof(md_addr_t));
	       /* and the data */
	       mem_strcpy(mem_access, mem, Write, sp, envp[i]);
	       sp += strlen(envp[i]) + 1;
	  }
     /* terminate the envp array with a NULL */
     mem_access(mem, Write, envp_addr + i*sizeof(md_addr_t),
		&null_ptr, sizeof(md_addr_t));

     /* did we tromp off the stop of the stack? */
     if (sp > ld_stack_base)
	  {
	       /* we did, indicate to the user that MD_MAX_ENVIRON must be increased,
		  alternatively, you can use a smaller environment, or fewer
		  command line arguments */
	       fatal("environment overflow, increase MD_MAX_ENVIRON in alpha.h");
	  }

     /* initialize the bottom of heap to top of data segment */
     ld_brk_point = ROUND_UP(ld_data_base + ld_data_size, MD_PAGE_SIZE);

     /* set initial minimum stack pointer value to initial stack value */
     ld_stack_min = regs->regs_R[MD_REG_SP];

     regs->regs_R[MD_REG_SP] = ld_environ_base;
     regs->regs_PC = ld_prog_entry;

     debug("ld_text_base: 0x%08x  ld_text_size: 0x%08x",
	   ld_text_base, ld_text_size);
     debug("ld_data_base: 0x%08x  ld_data_size: 0x%08x",
	   ld_data_base, ld_data_size);
     debug("ld_stack_base: 0x%08x  ld_stack_size: 0x%08x",
	   ld_stack_base, ld_stack_size);
     debug("ld_prog_entry: 0x%08x", ld_prog_entry);
}
