#include "program.h"
#include <glib.h>
#include "basicBlock.h"
#include "function.h"
#include "dfg.h"
#include "cfg.h"
#include "ssa.h"
#include "parse.h"
#include <assert.h>
#include <ctype.h>
//#include "controlDistribution.h"
#include "controlDependence.h"
#include "list.h"
#include "rhoFunctions.h"
#include "mem.h"
#include "phiFunctions.h"
#include "lcc.h"
#include  "dominance.h"
#include "epoch.h"

PProgram programNew() {
     PProgram n;
     n = (PProgram)calloc(sizeof(Program), 1);

     n->basicBlocksById = g_tree_new(int_compare); 
     n->basicBlocksByPC = g_tree_new(PC_ptr_compare); 
     n->instructionsByID = g_tree_new(int_compare);
     n->functionsByPC = g_tree_new(PC_ptr_compare);

     n->functionCallBlocksByPC = g_tree_new(PC_ptr_compare); 

     return n;
}

void programFree(PProgram p) {
     
     if (p->instructions) {
	  free(p->instructions);
     }
     g_tree_destroy(p->basicBlocksById);
     g_tree_destroy(p->basicBlocksByPC);
}



void programAddInst(PProgram prog, PInst inst)
{
     g_tree_insert(prog->instructionsByID, 
		   (void *)inst->id, inst);
}

PInst programGetInstByID(PProgram prog, int id)
{
     PInst r = g_tree_lookup(prog->instructionsByID, (void*)id);
     assert(r);
     return r;
}

void programAddFunction(PProgram prog, PFunction func)
{
     GList *l, *i;
     PBasicBlock block = NULL;
     PInst inst;
     PC *ppc;


     func->owner = prog;
     prog->functions = g_list_append(prog->functions, func);

     forEach(func->basicBlocks, l, block) {
	  g_tree_insert(prog->basicBlocksById, 
			(void *)block->id, block);
	  ppc = (PC *)calloc(sizeof(PC), 1);
	  *ppc = block->entrancePC;
	  g_tree_insert(prog->basicBlocksByPC, 
			ppc, block);
	  
	  forEach(block->insts, i, inst) {
	       programAddInst(prog, inst);
	  }
     }

     block = (PBasicBlock)g_list_nth(func->basicBlocks, 1)->data;
     
     ppc = (PC *)calloc(sizeof(PC), 1);
     *ppc = block->entrancePC;

     //     fprintf(stderr, "adding func @0x%llx\n", *ppc);
     func->pc = block->entrancePC;
     g_tree_insert(prog->functionsByPC, ppc, func);
}

PFunction programFindFunctionByPC(PProgram program, qword_t pc)
{
     PFunction f = g_tree_lookup(program->functionsByPC, &pc);

     return f;
}

PFunction programFindFunctionByName(PProgram program, char *funcName)
{
     GList *i;
     PFunction func;
     
     forEach(program->functions, i, func) {
	  if(!strcmp(func->name, funcName)) {
	       return func;
	  }
     }
     return NULL;
}

PBasicBlock programFindBlockByLabel(PProgram program, char* targetLabel) 
{
     GList *l1, *l2;
     PBasicBlock currentBlock;
     PFunction func;

     forEach(program->functions, l1,func) {
	  forEach(func->basicBlocks, l2, currentBlock) {
	       if(!strcmp(targetLabel, currentBlock->label)) {
		    return currentBlock;
	       }
	  }
     }
     return NULL;
}
PBasicBlock programFindBasicBlockByPC(PProgram program, PC pc)
{
     PBasicBlock r = g_tree_lookup(program->basicBlocksByPC, &pc);
     return r;
}

PBasicBlock programFindBasicBlockById(PProgram program, int id)
{
     PBasicBlock r = (PBasicBlock)g_tree_lookup(program->basicBlocksById, (void*)id);
     assert(r);
     return r;
}

SpecTableEntry * programGetCompileAction(PProgram program, char *function)
{
     GList *l;
     SpecTableEntry *e;

     forEach(program->compileSpec, l, e) {
	  if (!strcmp(function, e->functionName)){ 
	       return e;
	  }
     }
     return NULL;
}

void parseSpecLine(GList **l, char*buf)
{
     char *bufParts[30];
     int i;
     int p = 0;
     int len = strlen(buf);
     char *functionName;
     char *action;
     int actionNum;
     SpecTableEntry *e;
     nice_calloc(e, sizeof(SpecTableEntry));

     for(i = 0; i < 2; i++) {
	  bufParts[i] = 0;
     }

     p = 0;
     bufParts[p++] = buf;
     i = 0;
     for (i = 0; i < len;  i++) {
    
	  if (isspace(buf[i])) {
	       while(isspace(buf[i])) {
		    buf[i] = '\0';
		    i++;
	       }

	       bufParts[p++] = buf + i;
	  }

	  if (buf[i] == '\n') {
	       buf[i] = '\0';
	  }
     }
     
     i = 0;
     functionName = bufParts[i++];
     action = bufParts[i++];
     if (!strcmp(action, "compile")) {
	  e->action = DO_COMPILE;
     } else if (!strcmp(action, "stub")) {
	  e->action = DO_STUB;
     } else if (!strcmp(action, "skip")) {
	  e->action = DO_NOTHING;
     }
     
     for(; i < p; i++) {
	  if (strlen(bufParts[i]) ==0 ) {
	       continue;
	  }
	  if (!strcmp(bufParts[i], "isMain")) {
	       e->isMain = TRUE;
	       //	  } else if (!strcmp(action, "stub")) {
	  } else {
	       fprintf(stderr, "Illegal option (%s) is .compile file.\n", bufParts[i]);
	       assert(0);
	  }
     }
     
     fprintf(stderr, "%s::%s\n", functionName, action);
     e->functionName = strdup(functionName);
     
     *l = g_list_append(*l, e);

}

#define BUFSIZE 256
PProgram programLoad (FILE *f,  FILE *spec, int instCount)
{
     PProgram program;
     char buf[BUFSIZE];
     int count = 0;
     int instructionsParsed = 0;
     GList *i, *t;
     PFunction func;
     program = programNew();

     if (spec != NULL) {
	  while (fgets(buf, BUFSIZE, spec)) {
	       parseSpecLine(&program->compileSpec, buf);
	  }
     }
     
     info("Parsing...");
     while (fgets(buf, BUFSIZE, f)) {
	  parseLine(buf, 
		    program,
		    &instructionsParsed);
	  if (instructionsParsed > instCount && instCount != 0) {
	       fatal("Too many instructions in map file.  Out of space in instruction array.");
	  }
	  if ((count % 1000) == 0) {
	       progress(".");
	  }
	  count++;
     }
     if (instructionsParsed != instCount && instCount != 0) {
	  fatal("Too few instructions in map file. %d found; %d expected", instructionsParsed, instCount);
     }
     
     t = g_list_copy(program->functions);
     forEach(t,i, func) {
	  if (func->compileAction == DO_NOTHING) {
	       program->functions = g_list_remove(program->functions, func);
	  }
     }
     g_list_free(t);
     info("done.\n");

     return program;
}

void programCollectDfgStats(PProgram program, PDfgStats stats) 
{
     PFunction func;
     GList *l;
     memset(stats, 0, sizeof(DfgStats));

     forEach(program->functions, l, func) {
	  dfgCollectStats(func, stats);
     }
}


void fixUpProgram(PProgram program)
{
}

void programFixUpForExecution(PProgram program) 
{
     PFunction func;
     GList *l;
     

     forEach(program->functions, l, func) {
	  cfgBuild(func);
     }

     fixUpProgram(program);

     forEach(program->functions, l, func) {
	  functionPrint(stderr, func, cfgOptions.verbose);
     }

}

static void programCfgProcessing(PFunction func, bool_t doStub)
{

     //functionMarkEntries(func);
     cfgBuild(func);
     functionAddReturnBlock(func);     
     
     functionMakeLocalInvocation(func);
     if (doStub) {
	  functionMakeStub(func);
     }

     dominanceComputeDominatorTree(func, DOMINANCE_DIRECTION_PRE);
     //*     dominanceComputeDominanceFrontiers(func, DOMINANCE_DIRECTION_PRE);
     //*     dominanceComputeDominatorTree(func, DOMINANCE_DIRECTION_POST);
     //*     dominanceComputeDominanceFrontiers(func, DOMINANCE_DIRECTION_POST);
     cfgLabelBackEdges(func);
     //*     cfgComputeRechableSets(func, REACHABLE_BACKWARD, FALSE);

     cfgExpandLoops(func);


     dominanceTearDown(func);
     //*cfgReachableTearDown(func);
     dominanceComputeDominatorTree(func, DOMINANCE_DIRECTION_PRE);
     //*dominanceComputeDominanceFrontiers(func, DOMINANCE_DIRECTION_PRE);
     //*dominanceComputeDominatorTree(func, DOMINANCE_DIRECTION_POST);
     //*     dominanceComputeDominanceFrontiers(func, DOMINANCE_DIRECTION_POST);
     cfgLabelBackEdges(func);
     //*cfgComputeRechableSets(func, REACHABLE_BACKWARD, FALSE);

     cfgComputeRechableSets(func, REACHABLE_FORWARD, FALSE);
     //     functionAddMemNoopBlocks(func);

     cfgFindLoops(func);
     cfgAddExitBlocks(func);     

     cfgReachableTearDown(func);

     
     dominanceTearDown(func);
     cfgReachableTearDown(func);
     dominanceComputeDominatorTree(func, DOMINANCE_DIRECTION_PRE);
     //*dominanceComputeDominanceFrontiers(func, DOMINANCE_DIRECTION_PRE);
     //*dominanceComputeDominatorTree(func, DOMINANCE_DIRECTION_POST);
     dominanceComputeDominanceFrontiers(func, DOMINANCE_DIRECTION_POST);
     cfgLabelBackEdges(func);

     //*cfgComputeRechableSets(func, REACHABLE_BACKWARD, FALSE);

     cfgUpdateLoopBodies(func);

     //*strictControlDependenceCompute(func);  
     //*controlDependencesCompute(func);  
     
     //*controlDistributionInit(func);    
     //*controlDistrubutionOptimize(func);    
     functionRemoveAlignmentBlocks(func);	       

     
     functionSplitFunctionCallBlocks(func);
     functionAvoidEpochConfusion(func);


     if (!cfgOptions.supressCollapsing) {
	  functionCollapseBasicBlocks(func);
     }

     functionRemoveMarkedDeadBlocks(func);

     functionBuildEpochs(func);

}


/* These two function do most of the transformations.  There are two because
   there is a small amount of intraprocedural information (about what values a
   called function produces) needed */
static void programDfgProcessingOne(PFunction func)
{
     functionRemoveUnneededInstructions(func);

     if (cfgOptions.realFunctionCalls) {
	  functionInstallIndirectSendFunctionCalls(func);
     } else {
	  assert(0);
	  functionInstallFunctionCalls(func);
     }

     functionAddReturnSinks(func);

     functionFixReturns(func);

     functionFixConditionalMoves(func);     


     if (cfgOptions.serialize) {
	  functionSerializeInstructions(func);
     }

     // Here we add instructions to copy the entire register file from
     // block to block and then build the dfg to sort it all out.  We then
     // remove everything we don't need.
     functionInsertSpecialSources(func);
     rhoInsertRhoFunctions(func); 	  
     rhoInsertDummyCopies(func);

     assert(cfgOptions.performEpochNumbering);
     memInsertEpochAdvance(func);
     functionFixEntryEpochAdvances(func);
     if (!cfgOptions.doTokenMemory) {
	  functionAddMemNoOps(func);
     }

     rhoFixControlArgs(func);

     functionFixLocalInvocationInsts(func);
     //     functionPrint(stderr, func, cfgOptions.verbose);

     if (cfgOptions.doTokenMemory) {
	  functionTokenizeMemoryOps(func);
     }	  

     dfgBuild(func);

     //     phiSetPartners(func);
     //functionPrint(stderr, func, cfgOptions.verbose);
     
     rhoSplitInputs(func);
     rhoRemoveDummyCopies(func);

     // First remove the useless instructions.  This tells us what the outputs
     // of the functions really are.  Then prune the function calls and remove
     // useless instructions again.
     rhoRemoveUselessInstructions(func);
} 

static void programDfgProcessingTwo(PFunction func)
{
     functionPruneFunctionCallOutputs(func);
     rhoRemoveUselessInstructions(func);

     
     //phiCleanUpLoopPhiFunctions(func);
     //rhoCleanUpRhoFunctions(func);
     
     rhoOptimizeRhoFunctions(func);
     //     rhoFixOrphanedLoopRhoFuncs(func);
     rhoRemoveUselessInstructions(func);
	  
     //     functionRemoveUselessSpecialSources(func);
     if (!cfgOptions.supressEpochOptimize) {
	  epochOptimize(func);
     }

     
     rhoRemoveUselessInstructions(func);
     

     if (!cfgOptions.doTokenMemory) {
	  epochOptimizeCoordinators(func);
	  epochMarkUnneededCoordinators(func);
     }
     //cfgRemoveEmptyBlocks(func);
     rhoRemoveUselessInstructions(func);


     if (!cfgOptions.doTokenMemory) {
	  functionLabelCanonicalEpochAdvances(func);
     }
     
     if (cfgOptions.compressWaveAdvance) {
	  functionCompressWaveAdvance(func);
	  functionRemoveUnreachableInstructions(func);
     }
     functionBuildRippleChunks(func);

     // THis has to go late because we don't want to be numbering instructions
     // that get taken out.
     if (!cfgOptions.doTokenMemory) {
	  functionNumberMemOps(func);
	  functionSetInstructionEpochTags(func);
	  functionFixLocalInvocationEpoch(func);
	  functionEpochSanity(func);
     }

     functionComputeNoLocalPredecessorInsts(func);

     functionSetFirstInstBits(func);
     //     functionSetControllingCoordinator(func);

     functionSetInstructionLatencies(func);
     //	  controlDistributionAddDfgEdges(func); 
     functionCheckForNastiness(func);
     functionSanity(func);
}

void programPreTraceProcessing(PProgram program)
{
     PFunction func;
     GList *l;

     if (cfgOptions.onlyCFG) {
	  forEach(program->functions, l, func) {
	       programCfgProcessing(func,func->compileAction == DO_STUB);
	  }
     } else {
	  forEach(program->functions, l, func) {
	       programCfgProcessing(func,func->compileAction == DO_STUB);
	       programDfgProcessingOne(func);
	  }
	  forEach(program->functions, l, func) {
	       programDfgProcessingTwo(func);
	  }

	  //	  programRenumberBlocks(program);
	  //	  programCompressInstructionIDs(program);

     }
     
     if (cfgOptions.dumpFinal) {
	  forEach(program->functions, l, func) {
	       functionPrint(stderr, func, cfgOptions.verbose);
	  }
     }
}

void programPostTraceProcessing(PProgram program, bool_t analyzeOnly)
{
     GList *l1;
     PFunction func;

     forEach(program->functions, l1, func) {
	  if (!analyzeOnly) {
	       dfgDismantal(func);
	       dfgBuild(func);
	  }	   
	  
	  functionSanity(func);

     }
}

void programCompressInstructionIDs(PProgram program)
{
     GList *i, *j, *k;
     InstructionID cur = 0;
     PFunction func;
     PInst inst;
     PBasicBlock block;

     forEach(program->functions, i, func) {
	  forEach(func->basicBlocks, j, block) {
	       forEach(block->insts, k, inst) {
		    inst->id = cur++;
	       }
	  }
     }
}

void programRenumberBlocks(PProgram program) 
{
     GList *l, *i;
     PBasicBlock block;
     PFunction func;
     int c = 1;

     forEach(program->functions, i, func) {
	  forEach(func->basicBlocks, l, block) {
	       block->id = c++;
	  }
     }
}

void programPrintPerFunctionStats( PProgram program, FILE *f) 
{
     GList *l, *i;
     PFunction func;
     PDfgStats perFuncStats;

     nice_calloc(perFuncStats, sizeof(DfgStats));

     forEach(program->functions, i, func) {
	  memset(perFuncStats, 0, sizeof(DfgStats));
	  
	  dfgCollectStats(func, perFuncStats);	  
	  
	  fprintf(f, "%30s Overhead:  %4d; noopsRemoved %4d\n",func->name, perFuncStats->overheadInstructions, perFuncStats->extraInstructionCount + perFuncStats->branchCount);
     }
}

void programBuildRippleChunks(PProgram program)
{
     GList *l;
     PFunction func;
     forEach(program->functions, l, func) {
	  functionBuildRippleChunks(func);
     }
}
