#include "ptrSet.h"
#include "host.h"
#include "util.h"

static gint ptrCompare(gconstpointer a, gconstpointer b)
{
     unsigned int aa;
     unsigned int bb;

     aa = (unsigned int)a;
     bb = (unsigned int)b;
     
     if (aa < bb) {
	  return -1;
     } else if (aa > bb) {
	  return 1;
     } else {
	  return 0;
     }
}

PPtrSet ptrSetCopy(PPtrSet s)
{
     PPtrSet n = ptrSetNew();
     n->elements = g_list_copy(s->elements);
     return n;
}

PPtrSet ptrSetNew() 
{
     PPtrSet n;

     nice_calloc(n, sizeof(PtrSet));

     return n;
}

void ptrSetFree(PPtrSet s) 
{
     g_list_free(s->elements);
     free(s);
}

void ptrSetAdd(PPtrSet s, void *ptr)
{
     if (g_list_find_custom(s->elements, ptr, ptrCompare) == NULL) {
	  s->elements = g_list_insert_sorted(s->elements, ptr, ptrCompare);
     }
}

bool_t ptrSetElementOf(PPtrSet s, void *ptr) 
{
     if (g_list_find(s->elements, ptr)) {
	  return TRUE;
     } else {
	  return FALSE;
     }
}

void *ptrSetElement(PPtrSet s)
{
     if (s->elements) {
	  return s->elements->data;
     } else {
	  return NULL;
     }
}

/* This function has somewhat strange semantics.  Two consecutive calls with
   different indices will return different elements as long as the index is
   smaller than the cardinality of the set. */
void *ptrSetElementIndex(PPtrSet s, int i)
{
     if (s->elements) {
	  return g_list_nth_data(s->elements, i);
     } else {
	  return NULL;
     }
}

void ptrSetRemove(PPtrSet s, void *ptr)
{
     s->elements = g_list_remove(s->elements, ptr);
}

int ptrSetSize(PPtrSet s)
{
     return g_list_length(s->elements);
}

void ptrSetEmpty(PPtrSet s)
{
     g_list_free(s->elements);
     s->elements = NULL;
}

void ptrSetMinus(PPtrSet dst, PPtrSet a, PPtrSet b)
{
     PtrSetIndex i;
     void *e;

     if  (dst != a) {
	  ptrSetEmpty(dst);
	  ptrSetUnion(dst, dst, a);
     }

     forEachPtrElement(b, i, e) {
	  ptrSetRemove(a, e);
     }
     
}

bool_t ptrSetContains(PPtrSet a, PPtrSet b)
{
     PPtrSet t = ptrSetNew();
     bool_t r;

     ptrSetUnion(t, t, b);
     ptrSetMinus(t, t, a);
     
     if(ptrSetSize(t) == 0) {
	  r = TRUE;
     } else {
	  r = FALSE;
     }
     
     ptrSetFree(t);
     return r;
}

void ptrSetUnion(PPtrSet dst, PPtrSet a, PPtrSet b) 
{
     PtrSetIndex i;
     void *p;

     if (a != dst && b != dst) {
	  ptrSetEmpty(dst);
     }

     if (a != dst) {
	  forEachPtrElement(a, i, p) {
	       ptrSetAdd(dst, p);
	  }
     }

     if (b != dst) {
	  forEachPtrElement(b, i, p) {
	       ptrSetAdd(dst, p);
	  }
     }
}

void ptrSetIntersect(PPtrSet dst, PPtrSet a, PPtrSet b) 
{
     GList *i, *j, *t1, *t2;
     
     t1 = i = g_list_copy(a->elements);
     t2 = j = g_list_copy(b->elements);

     ptrSetEmpty(dst);

     while(i != NULL && j != NULL) {
	  if (i->data == j->data) {
	       ptrSetAdd(dst, i->data);
	       j = j->next;
	       i = i->next;
	  } else if (i->data < j->data) {
	       i = i->next;
	  } else {
	       j = j->next;
	  }
     }

     g_list_free(t1);
     g_list_free(t2);

}

bool_t ptrSetEqual(PPtrSet a, PPtrSet b)
{
     GList *i, *j;

     i = a->elements;
     j = b->elements;

     while(i != NULL && j != NULL) {
	  if (i->data != j->data) {
	       return FALSE;
	  }

	  i = i->next;
	  j = j->next;
     }

     if (i == j) { // They should both be NULL if the sets are equal
	  return TRUE;
     } else {
	  return FALSE;
     }
}
