#!/usr/bin/perl -w

package Drip::InstructionFormat;
use Exporter;
use Drip::Type qw( :ALL );
use Data::Dumper;
use strict ;
use Carp;
use Drip::IOString;
use Pebble::Constants; # Configures Bit::Vector

our @ISA = qw( Exporter );
our @EXPORT_OK = qw( SmallCount  InstructionID PC DripWSOpCode Immediate Wave Asid ThreadID
                     MemSequence RippleSize LatencyType InputArgType OutputArgType
                     CharString DripLink EncodedConsumer EncodedProducer EncodedOutput
                     EncodedInput EncodedInstruction EncodedFunction EncodedProgram
                     IF_VERSION
                   );

our %EXPORT_TAGS = ( ALL => \@EXPORT_OK );

Drip::Type::include 'Drip::InstructionFormatAux'; 
const &Version(), 'IF_VERSION', 10;

typedef struct('_DripLink',
               Field(MemSequence(), 'epochOrderPrev'),
               Field(MemSequence(), 'epochOrder'),
               Field(MemSequence(), 'epochOrderNext'),
               Field(RippleSize(),  'rippleSize'),
               Field(MemSequence(),  'lastStoreSequenceNumber'),
	       )         , 'DripLink';


typedef struct('_ENCODED_CONSUMER',
               Field(&InstructionID, 'inst'),
               Field(&InputArgType, 'type')  )             , 'EncodedConsumer';

typedef struct('_ENCODED_PRODUCER',
               Field(&InstructionID, 'inst'),
               Field(&OutputArgType, 'type') )             , 'EncodedProducer';

typedef &List(&EncodedConsumer)                            , 'EncodedOutput';
typedef &List(&EncodedProducer)                            , 'EncodedInput';

typedef struct('_ENCODED_INSTRUCTION',
               Field(&DripWSOpCode,         'opcode'),
               Field(&InstructionID,        'id'),                    #linker
               Field(&PC,                   'pc'),
               Field(&Immediate,            'immediate'),             #linker
               Field(&INT(LONG),            'extraInstructions'),
               Field(&INT(LONG),            'regNum'),
               Field(&INT(LONG),            'threadAdvanceAmount'),   #replaced waveAdvanceCount
	       Field(&INT(LONG),            'stubID'),                #linker
	       Field(&INT(LONG),            'isForLocalStub'),
	       Field(&INT(LONG),            'isZeroStore'),
	       Field(&INT(LONG),            'inAisZero'),
	       Field(&INT(LONG),            'inBisZero'),
	       Field(&INT(LONG),            'outAisZero'),
	       Field(&INT(LONG),            'reportValue'),
               Field(&INT(LONG),            'isPinned'),
               Field(&INT(LONG),            'pinRow'),
               Field(&INT(LONG),            'pinColumn'),
               Field(&INT(LONG),            'isForFunctionCall'),
               Field(&INT(LONG),            'waveAdvanceAmount'),
               Field(&INT(LONG),            'lineno'),
               Field(&DripLink,             'link'),
               Field(&LatencyType,          'latencyType'),
               Field(&List(&EncodedInput),  'inputs'),
               Field(&List(&EncodedOutput), 'outputs')
              )                                           , 'EncodedInstruction';

typedef struct('_ENCODED_FUNCTION',
               Field(&InstructionID, 'firstInstruction'),
               Field(&PC, 'pc'),
	       Field(&INT(LONG),            'stubID'),              #linker, stubID should be unique excetp for stub 0
	       Field(&PC,                   'localInvocationPC'),
	       Field(&INT(LONG),            'isNativeEntry'),
	       Field(&INT(LONG),            'isMain'),
               Field(&CharString, 'name')
              )                                           , 'EncodedFunction';
typedef &EncodedFunction, 'EncodedInstructionLabelDefinition';

typedef struct('_ENCODED_SHARED_ARC_TABLE_ENTRY',
	       Field(&CharString,              'name'),       # name of the shared arc
	       Field(&List(&EncodedConsumer), 'consumers'),  # all the inputs this arc leads to
	       Field(&List(&EncodedProducer), 'producers')   # all the outputs this arc comes from
	      ), 'EncodedSharedArcTableEntry';

typedef List(&EncodedSharedArcTableEntry), 'EncodedSharedArcTable';

#
# EncodedLabelUseAsArc
#
# Contains two lists, at most one of which should be non-empty.  The
# 'producers' list is a list of EncodedProducers specifying uses of the label
# by instructions producing data consumed by the labeled instruction.  The
# 'consumers' list is a list of EncodedConsumers specifying uses of the label
# by instructions consuming data produced by the labeled instruction.
#

typedef struct('_ENCODED_LABEL_USE_AS_ARC',
               Field(&List(&EncodedProducer),  'producers'), # one of these lists should 
               Field(&List(&EncodedConsumer),  'consumers'), # always be empty
              ), 'EncodedLabelUseAsArc';

#
# EncodedExternLabelTableEntry
#
# Each assembly unit produces one of these for every DOTEXTERN'd label.  It
# identifies the label by name, and contains two lists: the list of
# instructions using the labeled address as an immediate value, and the list of
# uses of the label to name communication arcs (see EncodedLabelUseAsArc, above).
#

typedef struct('_ENCODED_EXTERN_LABEL_TABLE_ENTRY',

               # The exported name of the external label
               # The linker should find exactly one match for this field
               # in exactly one of the object files being linked together.
	       Field(&CharString,                         'label'),

               # uses of the label as an immediate value
	       Field(&List(&InstructionID),               'immediateUses'),

               # uses of the label to identify an arc (e.g., ^label:2)
               # The index into field 'inputUses' is the absolute port number
               # of the labeled reference (i.e., the 2 in ^label:2).  It is
               # up to the linker to translate this number into the correct
               # InputArgType or OutputArgType using knowledge of the labeled
               # instruction.
	       Field(&List(&EncodedLabelUseAsArc),        'arcUses'),
	      ), 'EncodedExternLabelTableEntry';

typedef List(&EncodedExternLabelTableEntry), 'EncodedExternLabelTable';

typedef List(&InstructionID), 'ImmRelocationTable';

typedef struct('_ENCODED_PROGRAM',
               Field(&Version,                             'version'),
               Field(&ImmRelocationTable,                  'immRelocations'),
	       Field(&EncodedSharedArcTable,               'sharedArcs'),
	       Field(&EncodedExternLabelTable,             'externLabels'),
               Field(&List(&EncodedFunction),              'functions'), # this field doubles as the exported instruction label list
               Field(&List(&EncodedInstruction),           'instructions'),
              )                                          , 'EncodedProgram';


sub debugReadEncodedProgram
{
    EncodedProgram()->readInstance(\*STDIN);
}

sub debug {
    my $inst = List(MemSequence())->instantiate([MemSequence()->instantiate(500),
                                                 MemSequence()->instantiate(-1),
                                                 MemSequence()->instantiate(403)  ]);
    my $stream = new Drip::IOString;
    $inst->writeValue($stream);
    print "'$$stream'\n";
    my $instAnew = $inst->type->readInstance($stream);
    print "'$$stream'\n";
    print $_->value() . "\n" foreach( @{$instAnew->value} );

     $stream = new Drip::IOString;
     my $str = CharString()->instantiate([ UINT(BYTE)->instantiate(65),
                                           UINT(BYTE)->instantiate(66),
                                           UINT(BYTE)->instantiate(67),
                                           UINT(BYTE)->instantiate(68) ]);
    $str->writeValue($stream);
    print "$$stream\n";
    my $stranew = $str->type->readInstance($stream);
    print "$$stream\n";
    print join(", ", map { $_->value; } @{$stranew->value}), "\n";

#     UINT(LONGLONG)->generateCReadFn(\*STDOUT);
#     MemSequence()->generateCReadFn(\*STDOUT);
#     Link()->generateCReadFn(\*STDOUT);
#     print List(Link())->instanceDeclaration("mslist") . "\n";
#     MemSequenceList()->generateCReadFn(\*STDOUT);
#     List(Link())->generateCReadFn(\*STDOUT);

#    print Link()->basetype() . "\n";
#    print "$_\n" foreach(Drip::Type::allTypes());
#    Drip::Type::GenerateCDeclarations(\*STDOUT);
#    Drip::Type::GenerateCFnDefinitions(\*STDOUT);
}

1;
