#!/usr/bin/perl -w

package Drip::Type;
use English;
use Pebble::HashObject;
use Pebble::ArrayList;
use Drip::Field;
use Drip::Instance;
use Exporter;
use strict;
use Carp;
use constant BYTE     => 8;
use constant SHORT    => 16;
use constant LONG     => 32;
use constant LONGLONG => 64;
use Pebble::Stringy;

BEGIN {
    our @ISA = qw( Pebble::HashObject Exporter Pebble::Stringy );
    our %FIELDS = ( 
        '_defPkg'       => undef,
        '_defaultValue' => undef,
        'typestr' => undef
    );
    our @EXPORT_OK = qw( Field List include const typedef struct INT UINT ENUM BYTE SHORT LONG LONGLONG );
    our %EXPORT_TAGS = (
        ALL => [ @EXPORT_OK ],
        TYPES => [ qw( INT UINT ENUM List struct ) ],
        CONSTANTS => [ qw( BYTE LONG LONGLONG SHORT ) ],
    );

    __PACKAGE__->DECLARE_ACCESSORS;
}
our $AUTOLOAD;

my %TypeMap;
my %ConstMap;
my %IncludeMap;

sub allTypes { return values %TypeMap; }
sub getTypeMap { return \%TypeMap; }

sub typeDeclaration { croak("Unimplemented function!"); }
sub instanceDeclaration { croak("Unimplemented function!"); }

sub include {
    my $pkg = shift;
    my $callpkg = caller;
    carp "$callpkg including $pkg";
    eval "require $pkg;";
    if($EVAL_ERROR) { croak $EVAL_ERROR };
    $pkg->export_to_level(1, $pkg, ':ALL');
    $IncludeMap{$callpkg} = new Pebble::ArrayList unless defined $IncludeMap{$callpkg};
    $IncludeMap{$callpkg}->push($pkg);
}

sub const {
    my $type = shift;
    my $name = shift;
    my $val  = shift;
    my $callpkg = caller;
    my $fullname = "${callpkg}::${name}";
    
    croak("const $name already defined") if defined $ConstMap{$name};
    
    my $defdconst = declare(\%ConstMap, $type->instantiate($val), $fullname);
    no strict 'refs';
    *{$fullname} = sub() { 
	scalar @_ and carp("Too many parameters");
	my $retval = $defdconst; # This is here because of what appears to be a bug in perl 5.8
	return $retval;
    }
}

sub typedef {
    my $type = shift;
    my $name = shift;
    my $callpkg = caller;
    my $fullname = "${callpkg}::${name}";
    my $defdtype = declare(\%TypeMap, new Drip::Type::TypedefType($callpkg, $name, $type), $fullname);
    $defdtype->typestr($fullname);
#    print STDERR "Defined ", $defdtype->toString, ", or $fullname for long\n";
    no strict 'refs';
    *{$fullname} = sub () {
	scalar @_ and carp("Too many parameters");
#	print STDERR "returning $defdtype\n";
	my $retval = $defdtype; # This is here because of what appears to be a bug in perl 5.8
	return $retval;
    };
}

sub declare
{
    my $map = shift;
    my $type = shift;
    my $fullname = shift;
    croak("Type $fullname already defined") if exists $map->{$fullname};
    return  $map->{$fullname} = $type;
}

sub GenerateCFnDeclarations
{
    _GenerateCType(
        sub {
            my $fh = shift;
            my $type = shift;
            $fh->print($type->getCReadFnDecl() . ";\n\n") if defined $type->getCFnNameStem;
        }, @_);
}

sub GenerateCTypeDeclarations
{
    _GenerateCType(
        sub {
            my $fh = shift;
            my $type = shift;
            $fh->print("/**\n *\n * Declarations for type ", $type->toString, "\n *\n *\n */\n");
            $fh->print($type->typeDeclaration() . ";\n\n") if defined $type->typeDeclaration;
        }, @_);
}

sub GenerateCFnDefinitions
{
    _GenerateCType(
        sub {
            my $fh = shift;
            my $type = shift;
            $fh->print("/**\n * Definition of IO functions for $type\n */\n");
            $type->generateCReadFn($fh) if defined $type->getCFnNameStem;
            $fh->print("\n");
        }, @_);
}

sub _GenerateCType
{
    my $writerFn = shift;
    my $fh = shift;
    my $callpkg = shift;

    defined $fh or $fh = \*STDOUT;
    defined $callpkg or $callpkg = caller 1;
    my $depmap = { };

    foreach my $typename(keys %TypeMap) {
        if($typename =~ m/^${callpkg}::[^:]+$/) {
            _GenerateCType_aux($writerFn, $fh, $callpkg, $TypeMap{$typename}, $depmap);
        }
    }
}

sub _GenerateCType_aux
{
    my $writerFn = shift;
    my $fh = shift;
    my $callpkg = shift;
    my $type = shift;
    my $depmap = shift;
    my @deplist = $type->requiredTypes;

    foreach my $dep (@deplist) {
        unless(exists $depmap->{$dep->toString}) {
            _GenerateCType_aux($writerFn, $fh, $callpkg, $dep, $depmap);
        }
    }
    return if exists $depmap->{$type->toString};
    croak("Undefined defining package") unless defined $type->_defPkg;
    return if($type->_defPkg() ne $callpkg);
    $writerFn->($fh, $type);
    $depmap->{$type->toString} = 1;
}


sub GenerateCConstDefinitions
{
    my $fh = shift;
    my $callpkg = shift;

    defined $fh or $fh = \*STDOUT;
    defined $callpkg or $callpkg = caller;

    foreach my $fullname (keys %ConstMap) {
        if($fullname =~ m/^${callpkg}::[^:]+$/) {
            my $constant = $ConstMap{$fullname};
            my $constantName = substr($fullname, 2 + length $callpkg);
            $fh->print("const ", $constant->type->instanceDeclaration($constantName),
                       " = ", $constant->asStaticInitializer, ";\n\n");
        }
    }
}

sub _BaseFileForPkg
{
    my $retval = shift;
    $retval =~ s/::/\//g;
    return $retval;
}

sub _cHeaderFileForPkg
{
    return _BaseFileForPkg(@_) . ".h";
}

sub _cxxSourceFileForPkg
{
    return _BaseFileForPkg(@_) . ".cxx";
}

sub GenerateCIncludes
{
    my $fh = shift;
    my $callpkg = shift;

    defined $fh or $fh = \*STDOUT;
    defined $callpkg or $callpkg = caller;

    foreach my $incl ( @{$IncludeMap{$callpkg}} ) {
        carp("Including $incl");
        $fh->print("#include \"", _cHeaderFileForPkg($incl), "\"\n");
    }
}


sub Field($$) {
#    print STDERR join(", ", @_) . "\n";
    return new Drip::Field(@_);
}

sub initialize
{
    my $self = shift;
    my $pkg = shift;
    $self->_defPkg( $pkg );
    return $self;
}

sub instantiate
{ 
    my $self = shift;
    my $value = shift;
    $value = $self->defaultValue unless defined $value;
    if(ref($value) && UNIVERSAL::isa($value, 'Drip::Instance')) {
        $value = $value->value;
    }
    my $retval;
    eval { $retval = new Drip::Instance($self, $value); };
    if($@) { croak $@; }
    return $retval;
}    

sub defaultValue { my $self = shift; return $self->_defaultValue; }

sub getCReadFnName {
    my $self = shift;
    return "read" . $self->getCFnNameStem() . "Type";
}

sub getCReadFnDecl {
    my $self = shift;
    return "void " . $self->getCReadFnName() . "(FILE *file, " . $self->instanceDeclaration() . " *target)";
}

sub requiredTypes { return ( ); }

sub DESTROY { }

sub AUTOLOAD {
    no strict 'refs';
    my $fname = $AUTOLOAD;
    $fname =~ s/.*://;
    my $pkg = $AUTOLOAD;
    $pkg =~ s/(::[^:]*?)$//x;
    if($fname =~ /TypeMap/) { croak $AUTOLOAD; }
    croak("No such method $AUTOLOAD") unless $pkg eq __PACKAGE__;
    *$AUTOLOAD = sub {
        my $typestr = "$fname(" . join(", ", @_). ")";
        if(exists $TypeMap{$typestr}) {
            return $TypeMap{$typestr};
        }
        my $retval = $TypeMap{$typestr} = "Drip::Type::${fname}Type"->new(scalar caller, @_);
        $retval->typestr($typestr);
        return $retval;
    };
    goto &$AUTOLOAD;
}

use Drip::Type::INTType;
use Drip::Type::UINTType;
use Drip::Type::ENUMType;
use Drip::Type::structType;
use Drip::Type::TypedefType;
use Drip::Type::ListType;

1;
