#!/usr/bin/perl -w

package Drip::Type::INTType;
use strict;
use Bit::Vector::Overload;
use Carp;
use IO::Handle;

BEGIN {
    our @ISA = qw( Drip::Type );
    our %FIELDS = (
        '_precision' => undef,
        '_ctypestr' => undef,
        '_htonFn'   => undef,
        '_ntohFn'   => undef,
        '_cfnnamestem' => undef
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize {
    my $self = shift;
    my $pkg = shift;
    $self = $self->SUPER::initialize($pkg);
    my $precision = shift;

    $self->precision ( $precision );
    $self->_defaultValue( Bit::Vector->new_Dec($precision, 0) );
    return $self;
}

sub precision
{
    my $self = shift;
    if(scalar @_) {
        my $precision = shift;
        croak("precision must be a positive integer")
            unless ($precision == int($precision) && $precision > 0);
        if($precision & 7) {
            croak("Invalid precision $precision");
        }
        if($precision > Drip::Type::LONGLONG) { 
            croak("$precision exceeds maximum precision of " . Drip::Type::LONGLONG . "bits");
        }

        my ($ctypestr, $cfnnamestem, $htonFn, $ntohFn);
        if($precision    == Drip::Type::BYTE     )
        {
            $ctypestr = "int8_t";
            $cfnnamestem = "Int8";
            $htonFn = $ntohFn = "";
        }
        elsif($precision == Drip::Type::SHORT    )
        {
            $ctypestr = "int16_t"; 
            $cfnnamestem = "Int16";
            $htonFn = "htons";
            $ntohFn = "ntohs";
        }
        elsif($precision == Drip::Type::LONG     )
        {
            $ctypestr = "int32_t";
            $cfnnamestem = "Int32";
            $htonFn = "htonl";
            $ntohFn = "ntohl";
        }
        elsif($precision == Drip::Type::LONGLONG )
        {
            $ctypestr = "int64_t";
            $cfnnamestem = "Int64";
            $htonFn = "htonll";
            $ntohFn = "ntohll";
        }
        $self->_ctypestr($ctypestr);
        $self->_cfnnamestem($cfnnamestem);
        $self->_htonFn($htonFn);
        $self->_ntohFn($ntohFn);
        $self->_precision($precision);
    }
    return $self->_precision;
}

sub getCFnNameStem {
    my $self = shift;
    return $self->_cfnnamestem;
}

sub generateCReadFn {
    my $self = shift;
    my $fh = shift;
    my $type = shift;
    $type or $type = $self;
    my $creadfndecl = $type->getCReadFnDecl;
    my $instancedecl = $type->instanceDeclaration;
    my $ntohFn = $self->_ntohFn;
    
    print $fh <<"EOCReadFn";
$creadfndecl
{
    size_t amtread = fread( (void *)target, sizeof($instancedecl), 1, file);
    assert(amtread == 1);
    *target = ($instancedecl) $ntohFn(*target);
}
EOCReadFn
}

sub instanceDeclaration
{
    my $self = shift;
    my $name = shift;
    return $self->_ctypestr() . ($name ? " ${name}" : "");
}

sub typeDeclaration
{
    return undef;
}

sub asStaticInitializer
{
    my $type = shift;
    my $value = shift;
    return "0x$value";
}

# sub instanceDeclaration
# {
#     my $self = shift;
#     my $name = shift;
#     return $self->typeDeclaration($name);
# }

sub validate
{
    my $self = shift;
    my $value = shift;
    my $precision = $self->precision;
    croak("undefined value") unless defined $value;
    unless(UNIVERSAL::isa($value, 'Bit::Vector')) {
        $value = Bit::Vector->new_Dec($precision, $value);
    }
    unless($value->Size() == $precision) {
        $value = $value->Clone;
        $value->Resize($precision);
#        carp("Precision does not match; expected " . $self->precision() .
#             ", found " . $value->Size());
    }
    return $value;
}

sub valueAsBitVector {
    my ($self, $value) = @_;
#    print STDERR "Casting $value as type $self; was type ", ref($value) ? ref($value) : 'scalar', "\n";
    my $precision = $self->precision;
    unless (UNIVERSAL::isa($value, 'Bit::Vector')) {
        return Bit::Vector->new_Dec($precision, $value);
    }
    unless($value->Size() == $precision) {
        $value = $value->Clone;
        $value->Resize($precision);
    }
    return $value;
}

sub writeInstance
{
    my $self = shift;
    my $fh = shift;
    my $instance = shift;
    my $value = $instance->value;
    $value = $self->validate($value);
    $self->writeValue($fh, $instance->value);
}

sub writeValue
{
    my $val = $_[2];
    defined $val or $val = $_[0]->_defaultValue;
#    print STDERR (ref($val), , ": $val\n");
    $_[1]->print(scalar reverse(pack("C*", $val->Chunk_List_Read(8))));
}

sub readInstance
{
    my $self = shift;
    my $fh = shift;
    my $type = shift;
    defined $type or $type = $self;
    my $precision = $self->precision() >> 3;
    my $readstr;
    if($fh->read($readstr, $precision) != $precision) {
        croak("Read failed!");
    }
    my @readlist = reverse unpack("C*", $readstr);
#    carp join(", ", @readlist), "\n";
    my $bitvec = new Bit::Vector($self->precision);
    $bitvec->Chunk_List_Store(8, @readlist);
    return $type->instantiate($bitvec);
}

sub toString
{
    my $self = shift;
    return "INT(" . $self->precision() . ")";
}

1;
