#!/usr/bin/perl -w

package Drip::Type::ListType;
use strict;
use Carp;
use Pebble::ArrayList;

BEGIN {
    our @ISA = qw( Drip::Type );
    our %FIELDS = (
        '_itemType' => 'Drip::Type'
    );
    __PACKAGE__->DECLARE_ACCESSORS;
    Drip::Type->import qw( UINT LONG );
}

sub initialize {
    my $self = shift;
    my $pkg = shift;
    $self = $self->SUPER::initialize($pkg);
    my $itemType = shift;
    croak("Must specify type of list items") unless defined $itemType;
    $self->_itemType($itemType);
    $self->_defaultValue( new Pebble::ArrayList );
    return $self;
}

sub instanceDeclaration {
    my $self = shift;
    my $name = shift;
    $name or $name = "";
    return "struct _" . $self->_itemType->getCFnNameStem() . "List $name";
}

sub typeDeclaration {
    my $self = shift;
    my $name = shift;
    $name or $name  = "";
    return "struct _" . $self->_itemType->getCFnNameStem() ."List { " .
        UINT(LONG)->instanceDeclaration("length") . "; " .
        $self->_itemType->instanceDeclaration("*items") ."; } $name"
}

sub asStaticInitializer
{
    my $type = shift;
    my $value = shift;
    return "{ " . join(", ", map { $_->asStaticInitializer; } @{$value}) . " }";
}

sub getCFnNameStem
{
    my $self = shift;
    return $self->_itemType->getCFnNameStem() . "__List";
}

sub generateCReadFn
{
    my $self = shift;
    my $fh = shift;
    my $type = shift;
    $type or $type = $self;

    my $instancedecl = $type->instanceDeclaration;
    $fh->print($type->getCReadFnDecl . "\n{\n    " . UINT(LONG)->instanceDeclaration("i") . ";\n");
    $fh->print("    " . UINT(LONG)->getCReadFnName() . "(file, &target->length);\n");
    $fh->print("    if(target->length) {\n");
    $fh->print("        target->items = (".$self->_itemType->instanceDeclaration() ." *)calloc(sizeof(" . $self->_itemType->instanceDeclaration() .
               "), target->length);\n        assert(target->items);\n    }\n    else target->items = NULL;\n");
    $fh->print("    for(i = 0; i < target->length; i++) {\n");
    $fh->print("        " . $self->_itemType->getCReadFnName() . "(file, &target->items[i]);\n");
    $fh->print("    }\n}\n");
}

sub validate {
    my $self = shift;
    my $value = shift;
    croak("undefined value") unless defined $value;
    croak("Non-list structure") unless UNIVERSAL::isa($value, 'ARRAY');
    for(my $i = 0; $i < $#$value; $i++) {
        my $item = $value->[$i];
        croak("Undefined item in list") unless defined $item;
        croak(ref($item) or "non-ref scalar value", " is not a Drip::Instance")
            unless (UNIVERSAL::isa($item, 'Drip::Instance'));
        $item->value($self->_itemType->validate($item->value));
    }
    return $value;
}

sub writeInstance
{
    my $self = shift;
    my $fh = shift;
    my $instance = shift;
    my $value = $self->validate($instance->value);
    UINT(LONG)->instantiate(scalar(@$value))->writeValue($fh);
    foreach my $item (@{$value}) {
        croak("Unblessed scalar ", (ref($item) or $item)) unless UNIVERSAL::isa(ref($item), 'Drip::Instance');
        $item->writeValue($fh);
    }
}

sub writeValue
{
    my ($self, $fh, $value) = @_;
    my $itype = $self->_itemType;
    UINT(LONG)->writeValue($fh, UINT(LONG)->valueAsBitVector(scalar @{$value}));
    foreach my $item (@{$value}) {
        $itype->writeValue($fh, $item);
    }
}

sub readInstance
{
    my $self = shift;
    my $fh = shift;
    my $type = shift;
    $type or $type = $self;
    my $numitems = UINT(LONG)->readInstance($fh)->value;
    my @items;
    my $i;
#    carp("Reading a list of $numitems");
    for($i = UINT(LONG)->defaultValue->Clone; $i < $numitems; $i++) {
        push @items, $self->_itemType->readInstance($fh);
#        carp("Read item $i: ", $items[$#items]->value);
    }
#    carp("Read $i items");
    return $type->instantiate([@items]);
}

sub requiredTypes
{
    my $self = shift;
    return ( $self->_itemType, UINT(LONG) );
}

sub toString
{
    my $self = shift;
    return "List(" . $self->_itemType->toString() . ")";
}

1;
