#!/usr/bin/perl -w


package Drip::Type::structType;
use strict;
use Carp;

BEGIN {
    our @ISA = qw( Drip::Type );
    our %FIELDS = (
        'name'   => undef,
        'fields' => 'Pebble::ArrayList'
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize {
    my $self = shift;
    my $pkg = shift;
    $self->SUPER::initialize($pkg);
    if(defined $_[0] && !ref($_[0])) {
        $self->name(shift);
    } else {
        $self->name("");
    }
    my $defaults = $self->_defaultValue({ });
    $self->fields ( new Pebble::ArrayList );
    foreach my $field (@_) {
        $self->fields->push($field) ;
        $defaults->{$field->name} = $field->type->instantiate;
    }
    return $self;
}

sub getCFnNameStem
{
    my $self = shift;
    $self->name and return $self->name;
    return undef;
}

sub generateCReadFn
{
    my $self = shift;
    my $fh = shift;
    my $type = shift;
    $type or $type = $self;

    my $cfnnamestem = $type->getCFnNameStem;
    my $instancedecl = $type->instanceDeclaration;
    $fh->print($type->getCReadFnDecl() . "\n{\n");
    foreach my $field (@{$self->fields}) {
        my $fieldname = $field->name;
        my $fieldcreadfn = $field->type->getCReadFnName;
        $fh->print("    ${fieldcreadfn}(file, &target->${fieldname});\n");
    }
    $fh->print("}\n");
}

sub writeInstance
{
    my $self = shift;
    my $fh = shift;
    my $instance = shift;
    foreach my $field (@{$self->fields}) {
        $field->type->writeInstance($fh, $instance->value->{$field->name});
    }
}

sub writeValue
{
    my ($self, $fh, $value) = @_;
    my $field;
    foreach $field(@{$self->fields}) {
#           print STDERR "Writing ", $value->{$field->name}, " for field ", $field->name, "\n";
        $field->type->writeValue($fh, $value->{$field->name});
    }
}

sub readInstance
{
    my $self = shift;
    my $fh = shift;
    my $type = shift;
    $type or $type = $self;
    my $value = { };
    foreach my $field (@{$self->fields}) {
        $value->{$field->name} = $field->type->readInstance($fh);
    }
    return $type->instantiate($value);
}

sub typeDeclaration
{
    my $self = shift;
    my $name = shift;
    my $retval = "struct ". $self->name . " {\n";
    $retval .= "#ifdef __cplusplus\n";
    $retval .= "     explicit ". $self->name ."(const ".$self->name."&s) {\n";
#    $retval .= ("     "x2) . "if (s == NULL) return;\n";

    foreach my $field (@{$self->fields})
    {
        $retval .= ("    "x2) . $field->name . " = s.". $field->name . ";\n";
    }
    $retval .= "     ". "}\n";
    $retval .= "     ". $self->name ."() {}\n";
    $retval .= "#endif\n";
    foreach my $field (@{$self->fields})
    {
        $retval .= "    " . $field->type->instanceDeclaration($field->name) . ";\n";
    }
    $retval .= "}" .  (defined $name ? " $name" : "");
}

sub instanceDeclaration
{
    my $self = shift;
    my $iname = shift;
    $iname or $iname = "";
    if($self->name) {
        return "struct " . $self->name() . " $iname";
    }
    else {
        return $self->typeDeclaration($iname);
    }
}

sub validate
{
    my $self = $_[0];
    my $value = $_[1];
    croak("undefined value") unless defined $value;
#    croak(ref($value), " is a non-hash value for struct ",
#          ($self->name or "ANONYMOUS")) unless UNIVERSAL::isa($value, 'HASH');
    foreach my $field (@{$self->fields}) {
        my $fname = $field->name;
        my $ftype = $field->type;
        my $fval = $value->{$fname};
        unless(defined $fval) {
            ($fval = $value->{$fname} = $ftype->instantiate);
        }
###
### THE FOLLOWING CODE HAS BEEN COMMENTED OUT PURELY FOR SPEED; IT MAKES
### THE SYSTEM LESS ROBUST TO PROGRAMMING ERRORS IN THE ASSEMBLER
###
#         else {
#             croak("fields must be of type Drip::Instance")
#                 unless ref($fval) eq 'Drip::Instance';
# #            unless UNIVERSAL::isa($fval, 'Drip::Instance');
#             my $fvaltype = $fval->type;
#             croak("Expected field ", $fname, " to have type ", $ftype->toString,
#                   "; found type ", $fvaltype->toString) unless $fvaltype eq $ftype;
# #        $ftype->validate($fval->value);
#         }
    }
    return $value;
}

sub requiredTypes
{
    my $self = shift;
    my @retval = ();
    foreach my $field(@{$self->fields}) {
        push @retval, $field->type;
    }
    return @retval;
}

sub toString
{
    my $self = shift;
    return ("struct(" . ($self->name() ? "'". $self->name() . "', " : "") .
                  join(", ", @{$self->fields}) . ")");
}

1;
