#!/usr/bin/env perl

use warnings;
use strict;
use Pebble::HashObject;
use IO::File;

package LCTools::LockedFileQueue;
use Carp;
use Fcntl qw( :flock :seek );
use POSIX qw( O_RDWR O_CREAT );

BEGIN {
    our @ISA = qw( Pebble::HashObject );
    our %FIELDS = (
        'filehandle' => 'IO::File',
        'filename' => undef
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize {
    my ($self, $filename) = @_;
    my $fh = new IO::File;
    sysopen($fh, $filename, O_RDWR | O_CREAT)
        or croak("Couldn't open $filename for reading and writing");
    $self->filehandle($fh);
    $self->{_lockcnt} = 0;
    $self->filename($filename);
    return $self;
}

sub lock {
    my $self = $_[0];
    unless($self->{_lockcnt}) {
        until(flock($self->filehandle, LOCK_EX)) {
            carp("Failed to retrieve lock for file ", $self->filename, ": $!; retrying");
            sleep 1;
        }
    }
    ++$self->{_lockcnt};
    return 0;
}

sub unlock {
    my $self = $_[0];
    croak("Can't unlock an unheld lock") unless $self->{_lockcnt};
    unless(--$self->{_lockcnt}) {
        until(flock($self->filehandle, LOCK_UN)) {
            carp("Failed to release lock for file ", $self->filename, ": $!; retrying");
            sleep 1;
        }
    }
    return 0;
}

sub isLocked { return $_[0]->{_lockcnt} != 0; }

sub insertRecords
{
    my ($self, $index, @newrecs) = @_;
    $self->lock;
    my $fh = $self->filehandle;
    my @records = ();
    $fh->seek(0, SEEK_SET);
    push @records, (<$fh>);
    $fh->seek(0, SEEK_SET);
    $fh->print(@records[0..$index - 1])if $index > 0;
    $fh->print(map { "$_$/" } @newrecs);
    $fh->print(@records[$index..$#records]);
    $fh->flush;
    $self->unlock;
}

sub removeRecords
{
    my ($self, $index, $count) = @_;
    $self->lock;
    my $fh = $self->filehandle;
    $fh->seek(0, SEEK_SET);
    my $i;
    for($i = 0; $i < $index; $i++) {
        <$fh> or croak("Fewer than $index records");
    }
    my $pos = $fh->tell;
    my @retrecs = ();
    for($i = 0; $i < $count; $i++) {
        my $rec = <$fh>;
        defined $rec or croak("Fewer than ", $count+$index, " records");
        chomp($rec);
        push @retrecs, $rec;
    }
    my @tailrecs = ();
    push @tailrecs, (<$fh>);
    $fh->seek($pos, SEEK_SET);
    $fh->print(@tailrecs);
    $fh->truncate($fh->tell);
    $fh->flush;
    $self->unlock;
    return @retrecs;
}

sub enqueue {
    my ($self, @records) = @_;
    $self->insertRecords(0, @records);
}

#
# Returns undef if the queue was empty
#
sub dequeue {
    my ($self) = @_;
    $self->lock;
    my $fh = $self->filehandle;
    $fh->seek(0, SEEK_SET);
    $_ = undef;
    my ($pos, $oldpos, $lastline);
    $oldpos = 0;
    $lastline = undef;
    $pos = $fh->tell;
    while(<$fh>) {
        $oldpos = $pos;
        $pos = $fh->tell;
        $lastline = $_;
#        print STDERR "Read $lastline\n";
    }
#    print STDERR "Mysteriously ignored '$_'\n";
    chomp($lastline) if defined $lastline;
    $fh->truncate($oldpos);
    $fh->flush;
    $self->unlock;
    return $lastline;
}

sub debug_enq {
    my $q = __PACKAGE__->new("q.txt");
    for(my $i = 0; $i < 10; $i++) {
        sleep 1;
        $q->enqueue($i);
    }
}

sub debug_deq {
    my $q = __PACKAGE__->new("q.txt");
    for(;;) {
        sleep 1;
        my $val = undef;
        until(defined ($val = $q->dequeue)) {
            sleep 1;
        }
        print "Retrieved $val\n";
    }
}

1;
