#!/usr/bin/perl -w

package Pebble::Generator::ExecFormatGenerator;

use strict;
use Carp;
use Pebble::HashObject;
use Pebble::Visitor;
use Pebble::Generator::IndexingFlattener;
use Pebble::ArrayList;
use Drip::Type qw( :ALL );
use Drip::InstructionFormat qw( :ALL );

BEGIN {
    our @ISA = qw( Pebble::HashObject Pebble::Visitor );
    our %FIELDS = (
        'outstream' => undef,
        'encodedinstrs' => 'Pebble::ArrayList'
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize {
    my $self = shift;
    $self->outstream(shift);
    $self->encodedinstrs(new Pebble::ArrayList);
    return $self;
}


sub visitAssemblyUnit
{
    my $self = shift;
    my $asmUnit = shift;
    my @sections = $asmUnit->sections;
    my @labels = $asmUnit->labels;

    # flatten sections
    $asmUnit->visit(new Pebble::Generator::IndexingFlattener);
    $self->SUPER::visitAssemblyUnit($asmUnit);

    # all labels must be defined, as this is an executable
    foreach my $label (@labels) {
        croak("Undefined label ", $label->toString) unless defined $label->getInstruction;
    }

    # The export table serves to identify functions in the Kahuna ExecFormat.
    my $fnlist = $self->generateFunctionList($asmUnit->_exportedSymbols);

    my $executable = EncodedProgram()->instantiate({
        version      => (defined $asmUnit->version() ? IF_VERSION->type->instantiate($asmUnit->version)
                                                     : IF_VERSION),
        functions    => $fnlist,
        instructions => &List(&EncodedInstruction)->instantiate($self->encodedinstrs)
    });

    $executable->writeValue($self->outstream);
}

sub visitInstrInstance
{
    my $generator = shift;
    my $instr = shift;
    my $instrdef = $instr->instrdef;
#    print STDERR "Instr Instance Index: ", $instr->getCodeIndex, "\n";

    my $encinstr;
    $generator->encodedinstrs->push( $encinstr = EncodedInstruction()->instantiate({
        opcode                => DripWSOpCode()->instantiate($instrdef->opcode),
        id                      => InstructionID()->instantiate($instr->getCodeIndex),
        immediate               => Immediate()->instantiate( $instrdef->HasImmediate ?
                                               $instr->operands->[$#{$instr->operands}]->value : 0 ),
        link                    => $generator->generateLink($instr->waveorder),
        inputs                  => $generator->generateEncodedInputs($instr),
        outputs                 => $generator->generateEncodedOutputs($instr)
    }) );

    $generator->applyAttributes($encinstr, $instr->attributes);
}

sub applyAttributes
{
    my $generator   = shift;
    my $encinstance = shift;
    my $attributes  = shift;

    foreach my $field( @{$encinstance->type->fields} ) {
        if( defined $attributes->{$field->name} ) {
            $encinstance->value->{$field->name} =
                $field->type->instantiate($attributes->{$field->name});
        }
    }
    $encinstance->validate;
}


sub generateLink
{
    my $generator = shift;
    my $link = shift;
    return DripLink()->instantiate() unless(defined $link);
    return DripLink()->instantiate({epochOrderPrev => MemSequence()->instantiate($link->prev),
                                    epochOrder     => MemSequence()->instantiate($link->curr),
                                    epochOrderNext => MemSequence()->instantiate($link->next),
                                    rippleSize     => RippleSize()->instantiate($link->ripple)});
}

sub generateEncodedInputs
{
    my $generator = shift;
    my $instr = shift;
    my $instrdef = $instr->instrdef;
    my @operands = @{$instr->operands};
    my $encodedInputs = [];

    for(my $i = $instrdef->numOutputArcs; $i < $instrdef->numArcOperands; $i++) {
        my $operand = $operands[$i];
        push @$encodedInputs, $generator->generateEncodedProducers($operand->producers);
    }
    return &List(&EncodedInput)->instantiate($encodedInputs);
}

sub generateEncodedOutputs
{
    my $generator= shift;
    my $instr = shift;
    my $instrdef = $instr->instrdef;
    my @operands = @{$instr->operands};
    my $encodedOutputs = [];

    for(my $i = 0; $i < $instrdef->numOutputArcs; $i++) {
        my $operand = $operands[$i];
        push @$encodedOutputs, $generator->generateEncodedConsumers($operand->consumers);
    }
    return &List(&EncodedOutput())->instantiate($encodedOutputs);
}

sub generateEncodedProducers
{
    my ($generator, @producers) = @_;
    my $encProducers = [ ];

    foreach my $producer (@producers) {
        push @$encProducers, EncodedProducer()->instantiate({
             inst => InstructionID()->instantiate($producer->getCodeIndex),
             type => OutputArgType()->instantiate($producer->getPortType)
        });
    }
    return EncodedInput()->instantiate($encProducers);
}


sub generateEncodedConsumers
{
    my ($generator, @consumers) = @_;
    my $encConsumers = [ ];

    foreach my $consumer (@consumers) {
        push @$encConsumers, EncodedConsumer()->instantiate({
            inst => InstructionID()->instantiate($consumer->getCodeIndex),
            type => InputArgType()->instantiate($consumer->getPortType)
            });
    }
    return EncodedOutput()->instantiate($encConsumers);
}

sub generateFunctionList
{
    my $generator = shift;
    my $exportedSymbols = shift;
    my @fnentries = ();

    while(my ($name, $labelref) = each %$exportedSymbols)
    {
        my $fnentry;
        push @fnentries, $fnentry = EncodedFunction()->instantiate({
            firstInstruction => InstructionID()->instantiate($labelref->getInstruction->getCodeIndex),
            name => CharString()->instantiate([map { UINT(BYTE)->instantiate($_); } unpack("C*", $name . "\0")])
            });
        $generator->applyAttributes($fnentry, $labelref->attributes);
    }
    return List(EncodedFunction())->instantiate([@fnentries]);
}

sub getOutputExtension { return ".ws"; }

1;
