#!/usr/bin/perl -w

package Pebble::Generator::FastExecFormatGenerator;

use strict;
use Carp;
use Pebble::HashObject;
use Pebble::Visitor;
use Pebble::Generator::IndexingFlattener;
use Pebble::ArrayList;
use Drip::Type qw( :ALL );
use Drip::InstructionFormat qw(IF_VERSION) ;#qw( :ALL );

BEGIN {
    our @ISA = qw( Pebble::HashObject Pebble::Visitor );
    our %FIELDS = (
        'outstream' => undef,
        'encodedinstrs' => 'Pebble::ArrayList',
        'reloctable' => 'Pebble::ArrayList'
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

use constant ENCODED_INSTRUCTION_TYPE => Drip::InstructionFormat::EncodedInstruction();
use constant ENCODED_FUNCTION_TYPE => Drip::InstructionFormat::EncodedFunction();
use constant ENCODED_EXECUTABLE_TYPE => Drip::InstructionFormat::EncodedProgram();
use constant MEMORDER_TYPE => Drip::InstructionFormat::MemSequence();
use constant RIPPLE_TYPE => Drip::InstructionFormat::RippleSize();
use constant OPCODE_TYPE => Drip::InstructionFormat::DripWSOpCode();
use constant INSTR_ID_TYPE => Drip::InstructionFormat::InstructionID();
use constant IMM_TYPE => Drip::InstructionFormat::Immediate();
use constant OUTPUT_ARG_TYPE => Drip::InstructionFormat::OutputArgType();
use constant INPUT_ARG_TYPE => Drip::InstructionFormat::InputArgType();
use constant CHAR_STR_TYPE => Drip::InstructionFormat::CharString();
use constant CHAR_TYPE => CHAR_STR_TYPE->_itemType;

use constant MEMORDER_ZERO => MEMORDER_TYPE->valueAsBitVector(0);
use constant RIPPLE_ZERO => RIPPLE_TYPE->valueAsBitVector(0);

sub initialize {
    my $self = shift;
    $self->outstream(shift);
    $self->encodedinstrs(new Pebble::ArrayList);
    $self->reloctable(new Pebble::ArrayList);
    return $self;
}

sub _makeApplyAttributesToTypeFn
{
    my $type = shift;
    my $typename = shift;
    my @fieldtypes = map { $_->type; } @{$type->fields};
    my $fnstr = "package " . __PACKAGE__ . "; sub applyAttributesTo$typename {\n    my (\$generator, \$instance, \$attributes, \@rest) = \@_;\n";
    my $i = 0;
    foreach my $field (@{$type->fields}) {
        my $fieldname = $field->name;
        my $typestr = $field->type->typestr;
        $fnstr .= "    defined \$attributes->{'$fieldname'} and (\$instance->{'$fieldname'} = Drip::Type::getTypeMap->{'$typestr'}->valueAsBitVector(\$attributes->{'$fieldname'}));\n";
        ++$i;
    }
    $fnstr .= "}\n";
#    print STDERR $fnstr;
    eval $fnstr;
    if($@) { die $@; }
}

BEGIN {
    _makeApplyAttributesToTypeFn(ENCODED_INSTRUCTION_TYPE, 'EncodedInstruction');
    _makeApplyAttributesToTypeFn(ENCODED_FUNCTION_TYPE, 'EncodedFunction');
}

sub visitAssemblyUnit
{
    my $self = shift;
    my $asmUnit = shift;
    my @sections = $asmUnit->sections;
    my @labels = $asmUnit->labels;

#    print STDERR "Flattening: ", time, "\n";
    # flatten sections
    $asmUnit->visit(new Pebble::Generator::IndexingFlattener);
#    print STDERR "Generating instructions:", time, "\n";
    $self->SUPER::visitAssemblyUnit($asmUnit);

    # all labels must be fully specified, meaning defined or extern;
    foreach my $label (@labels) {
        croak("Undefined label ", $label->toString) unless $label->isSpecified;
    }

#    print STDERR "Generating fn list: ", time, "\n";
    # The export table serves to identify functions in the Kahuna ExecFormat.
    my $fnlist = $self->generateFunctionList($asmUnit->_exportedSymbols);

    
    my $executable ={ 
        version      => (defined $asmUnit->version() ? IF_VERSION->type->valueAsBitVector($asmUnit->version)
			 : IF_VERSION->value),
            functions    => $fnlist,
            instructions => $self->encodedinstrs,
	    externLabels => $self->generateExternsList($asmUnit->_externSymbols),
            sharedArcs   => $self->generateSharedArcList($asmUnit->getSharedArcs),
            immRelocations=>$self->reloctable
        };

#    eval {
#    print STDERR "Writing: ", time, "\n";
    ENCODED_EXECUTABLE_TYPE->writeValue($self->outstream, $executable);
#    print STDERR "Done: ", time, "\n";
#    }; if($@) { 
#        print STDERR Dumper($executable);
#        carp($@);
#    }
}

sub visitInstrInstance
{
    my $generator = shift;
    my $instr = shift;
    my $instrdef = $instr->instrdef;
#    print STDERR "Instr Instance Index: ", $instr->getCodeIndex, "\n";

    my $encinstr;
    $generator->encodedinstrs->push( $encinstr = {
        opcode                => OPCODE_TYPE->valueAsBitVector($instrdef->opcode),
        id                      => INSTR_ID_TYPE->valueAsBitVector($instr->getCodeIndex),
        immediate               => IMM_TYPE->valueAsBitVector($instrdef->HasImmediate ? $instr->operands->[$#{$instr->operands}]->value : 0),
        link                    => $generator->generateLink($instr->waveorder),
        inputs                  => $generator->generateEncodedInputs($instr),
        outputs                 => $generator->generateEncodedOutputs($instr)
    });

    if( $instrdef->HasImmediate &&
        ref($instr->operands->[$#{$instr->operands}]) eq "Pebble::Immediate::LabelRef" )
    {
        $generator->addImmRelocEntry($instr->getCodeIndex);
    }

    $generator->applyAttributesToEncodedInstruction($encinstr, $instr->attributes);
}

sub generateLink
{
    my ($generator, $link, @rest) = @_;
    return { } unless(defined $link);
    return {epochOrderPrev => MEMORDER_TYPE->valueAsBitVector($link->prev),
            epochOrder     => MEMORDER_TYPE->valueAsBitVector($link->curr),
            epochOrderNext => MEMORDER_TYPE->valueAsBitVector($link->next),
            rippleSize     => RIPPLE_TYPE->valueAsBitVector($link->ripple)};
}

sub generateEncodedInputs
{
    my $generator = shift;
    my $instr = shift;
    my $instrdef = $instr->instrdef;
    my @operands = @{$instr->operands};
    my $encodedInputs = [];

    for(my $i = $instrdef->numOutputArcs; $i < $instrdef->numArcOperands; $i++) {
        my $operand = $operands[$i];
        push @$encodedInputs, $generator->generateEncodedProducers($operand->producers);
    }
    return $encodedInputs;
}

sub generateEncodedOutputs
{
    my $generator= shift;
    my $instr = shift;
    my $instrdef = $instr->instrdef;
    my @operands = @{$instr->operands};
    my $encodedOutputs = [];

    for(my $i = 0; $i < $instrdef->numOutputArcs; $i++) {
        my $operand = $operands[$i];
        push @$encodedOutputs, $generator->generateEncodedConsumers($operand->consumers);
    }
    return $encodedOutputs;
}

sub generateEncodedProducers
{
    my ($generator, @producers) = @_;
    my $encProducers = [ ];

    foreach my $producer (@producers) {
        push @$encProducers, {
            inst => INSTR_ID_TYPE->valueAsBitVector($producer->getCodeIndex),
            type => OUTPUT_ARG_TYPE->validate($producer->getPortType)
            };
    }
    return $encProducers;
}


sub generateEncodedConsumers
{
    my ($generator, @consumers) = @_;
    my $encConsumers = [ ];

    foreach my $consumer (@consumers) {
        push @$encConsumers, {
            inst => INSTR_ID_TYPE->valueAsBitVector($consumer->getCodeIndex),
            type => INPUT_ARG_TYPE->validate($consumer->getPortType)
            };
    }
    return $encConsumers;
}

sub generateFunctionList
{
    my $generator = shift;
    my $exportedSymbols = shift;
    my $fnentries = [];

    while(my ($name, $labelref) = each %$exportedSymbols)
    {
        my $fnentry;
        push @{$fnentries}, $fnentry = {
            firstInstruction => INSTR_ID_TYPE->valueAsBitVector($labelref->getInstruction->getCodeIndex),
            name => $generator->encodedCharString($name)
            };
        $generator->applyAttributesToEncodedFunction($fnentry, $labelref->attributes);
    }
    return $fnentries;
}

sub encodedCharString
{
    my ($self, $name) = @_;
    return [map { CHAR_TYPE->valueAsBitVector($_) } unpack("C*", $name . "\0")];
}

sub generateSharedArcList
{
    my $generator = shift;
    my $descSharedArcs = [ ];
    foreach my $arc (@_) {
        push @{$descSharedArcs}, {
            name => $generator->encodedCharString($arc->name),
            consumers => $generator->generateEncodedConsumers($arc->consumers),
            producers => $generator->generateEncodedProducers($arc->producers)
            };
    }
    return $descSharedArcs;
}

sub generateExternsList {
    my ($generator, $externSymbolTable) = @_;
    my $externsList = new Pebble::ArrayList;

    foreach my $name (keys %{$externSymbolTable}) {
        my $label = $externSymbolTable->{$name};

#        defined $_ or croak($label->name) foreach $label->arcs;
        $externsList->push({
            label         => $generator->encodedCharString($name),
            immediateUses =>
                [ map {
                    INSTR_ID_TYPE->valueAsBitVector($_->getCodeIndex);
                } $label->nonArcReferences ],
            arcUses       =>
                [ map { defined $_ ?
                        { producers => $generator->generateEncodedProducers($_->producers),
                          consumers => $generator->generateEncodedConsumers($_->consumers) }
                        : { producers => [ ], consumers => [ ] };
                } $label->arcs ]
            });
    }
    return $externsList;
}

sub addImmRelocEntry
{
    my ($generator, $index) = @_;
    $generator->reloctable->push(INSTR_ID_TYPE->valueAsBitVector($index));
#   print STDERR "Adding index $index to relocation table\n";
}

sub getOutputExtension { return ".ws"; }

sub DESTROY { }
sub AUTOLOAD {
    croak;
}

1;
