#!/usr/bin/perl -w

package Pebble::Generator::TextGenerator;

use strict;
use Carp;
use Pebble::HashObject;
use Pebble::ArrayList;
use Pebble::Visitor;
use Pebble::Generator::IndexingFlattener;
use IO::Handle;


BEGIN {
    our @ISA = qw( Pebble::HashObject Pebble::Visitor );
    our %FIELDS = (
        'outstream' => undef
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}

sub initialize {
    my $self = shift;
    $self->outstream(shift);
    return $self;
}

sub visitAssemblyUnit
{
    my $self = shift;
    my $asmUnit = shift;
    my @sections = $asmUnit->sections;
    my @labels = $asmUnit->labels;
    my @undefinedlabels = ();
    
    # flatten sections
    $asmUnit->visit($self->getFlattener);

    # generate imports table
    foreach my $label (@labels) {
        push @undefinedlabels, $label unless defined $label->getInstruction;
    }
    $self->putUndefTable(@undefinedlabels);

    # generate exports table
    $self->putExportTable($asmUnit->_exportedSymbols);

     #     generate relocation table
#    foreach my $section (@sections) {
#    }

    $self->SUPER::visitAssemblyUnit($asmUnit);
}

sub visitSection
{
    my $self = shift;
    my $section = shift;
    $self->outstream->print("SECTION " . $section->name . "\n");
    $self->SUPER::visitSection($section);
}

sub stringOfOperandPorts
{
    return join(", ",
                map({ "^" . $_->getCodeIndex() . ":" . $_->instrPort(); } @_));
}

sub putUndefTable {
    my $self = shift;
    my $os = $self->outstream;

    $os->print("UNDEFINED LABELS\n");
    foreach my $label (@_) {
        $os->print($label->name, " => ", stringOfOperandPorts($label->nonArcReferences), "\n" );
        my $port = 0;
        foreach my $arc ( $label->arcs ) {
            next unless defined $arc;
            $os->print("^", $label->name, ":", $port, " => ",
                       stringOfOperandPorts($arc->referingInstrInstances), "\n");
            $port++;
        }
    }
    $os->print("\n");
}

sub putExportTable { $_[0]->putSymbolTable("EXPORT SYMBOLS", @_[1..$#_]); }

sub putSymbolTable
{
    my $self = shift;
    my $name = shift;
    my $exportMap = shift;
    my $os = $self->outstream;

    $os->print("$name TABLE\n");
    while(my ($name, $labelref) = each %$exportMap)
    {
        my $instr = $labelref->getInstruction;
        $os->print("$name (");
        foreach my $attr (keys %{$labelref->attributes}) {
            my $val = $labelref->attributes->{$attr};
            $os->print("$attr = $val,");
        }
        $os->print(") => ");
        if($instr) {
            $os->print($instr->section->name() . ":" . $instr->getCodeIndex());
        } else {
            $os->print(" ????");
        }
        $os->print("\n");
    }
    $os->print("\n");
}

sub visitInstrInstance
{
    my $self = shift;
    my $os = $self->outstream;
    my $instr = shift;

    $os->printf("%5d: %-10s (%3d) ", $instr->getCodeIndex,
                $instr->instrdef->opname, $instr->instrdef->opcode);
    my @operands = @{$instr->operands};
    if($instr->instrdef->HasImmediate) {
        $os->print(" #0x", $operands[$#operands]->value, " ");
    } else { $os->print(" " x 20); }
    for(my $i = 0; $i < $instr->instrdef->numOutputArcs; $i++) {
        &doPrintTargets($os, 'consumers', shift(@operands)->arcs);
    }
    for(my $i = 0; $i < $instr->instrdef->numInputArcs; $i++) {
        &doPrintTargets($os, 'producers', shift(@operands)->arcs);
    }
    if($instr->instrdef->isMemoryInstr) {
        $os->print($instr->waveorder->toString);
    }
    $os->print(" [ ");
    foreach my $attrkey(keys %{$instr->attributes}) {
        $os->print("$attrkey => ", $instr->attributes->{$attrkey}, " ");
    }
    $os->print("] \n");
}

sub doPrintTargets
{
    my $os = shift;
    my $kind = shift;
    $os->print(" { ");
    $os->print(join ", ", map { stringOfOperandPorts($_->$kind); } @_);
    $os->print(" }, ");
}

sub getFlattener { return new Pebble::Generator::IndexingFlattener; }

sub getOutputExtension { return ".txt"; }

1;
