#!/usr/bin/perl -w

use strict;
use Pebble::HashObject;
use Pebble::NamedThing;
use Pebble::Arc;
use Pebble::InstrLabel;
use Carp;


#
# This class is a fully implicit namespace
#
package Pebble::Namespace;
use constant FALSE => undef;
use constant TRUE  => 1;

BEGIN {
    our @ISA = qw( Pebble::HashObject );
    our %FIELDS = (
        'name'      =>  undef,
        '_parent'    =>  undef,
        '_kids'      => 'HASH',
        '_namespaces' => 'HASH',
        '_asmunit'    => 'Pebble::AssemblyUnit'
    );
    __PACKAGE__->DECLARE_ACCESSORS;
}
        
sub initialize
{
    my $self = shift;
    my $nsparent = shift;
    my $name = shift;
    my $asmunit = shift;
    Carp::croak("No asmunit") unless defined $asmunit;
    scalar @_ and carp("Too many parameters");
    $self->name($name);
    $self->_parent($nsparent);
    $self->_kids({ });
    $self->_namespaces({ });
    $self->_asmunit($asmunit);

    return $self;
}

sub fullname {
    my $self = shift;
    return ($self->_parent() ? $self->_parent->fullname() : "") .
           ($self->name() ? $self->name() : $self) . "::";
}

sub specialize
{
    my $self = shift;
    my $newkind = shift;

    croak("Can only specialize to subclasses of " . __PACKAGE__)
        unless UNIVERSAL::isa($newkind, __PACKAGE__);

    bless $self, $newkind;
    $self->doSpecialize;
}

sub isSpecialized
{
    my $self = shift;
    return ref($self) ne __PACKAGE__;
}

sub importSymbols
{
    my $self = shift;
    if(ref($_[0]) eq "HASH")
    {
        my $imports = shift;
        @_ = %$imports;
    }
    while(scalar @_) {
        my $name = shift;
        my $thing = shift;
        $self->putByName($name, $thing, 1);
    }
}

sub getNamedThingByNameLocally
{
    my $self = shift;
    my $name = shift;
#    $name or Carp::carp("Undefined name!");
    return $self->_kids->{$name};
}

#
# Searches for the NamedThing with the given name in this, or the parent
# Namespace.  Subclasses requiring explicit name importing will override
# this with ExplicitImportNamespace::getNamedThingByName .
#
sub getNamedThingByName
{
    my $self = shift;
    my $name = shift;

    $self->getNamedThingByNameLocally($name) or
        $self->_parent and $self->_parent->getNamedThingByName($name);
}

#
# Puts the given object into the namespace with the given name,
# or dies if something in the namespace already has that name.
# Things in the namespace are actually stored as referents in
# NamedThing objects, which also have "declared" bits.
#
sub putByName
{
    my $self = shift;
    my $name = shift;
    my $thing = shift;
    my $declared = 0;
    scalar @_ and $declared = shift;
    $name or Carp::croak("The name of the thing to put must not be undef!");
    $thing or Carp::croak("The thing to put must not be undef!  Perhaps an unblessed HASH?");

    $self->_kids->{$name} and Carp::croak("Name $name already in use");
    $self->_kids->{$name} = new Pebble::NamedThing($thing, $declared);

    return $thing;
}

#
# Return the object with the given name, or undef if there is no such thing.
# Works by calling getNamedThingByName.
#
sub getByName
{
    my $self = shift;
    my $name = shift;
    my $namedthing = $self->getNamedThingByName($name);
    return $namedthing->referent if $namedthing;
    return undef;
}

sub getByNameLocally
{
    my $self = shift;
    my $name = shift;
    my $namedthing = $self->getNamedThingByNameLocally($name);
    return $namedthing->referent if $namedthing;
    return undef;
}

#
# Return the object in this namespace with the given name, dying if the type is wrong,
# and inserting a new object into the namespace if one did not already exist.
#
sub getByNameTyped
{
    my $self = shift;
    my $name = shift;
    my $type = shift;
    my $accessor = shift;
    $accessor or $accessor = \&{__PACKAGE__ . "::getByName"};
    scalar @_ and carp("Too many parameters");

    my $retval = $accessor->($self, $name);
    unless (defined $retval) {
        $retval = $self->putByName($name, new $type($self->fullname() . $name, $self->_asmunit));
    }
    if(ref($retval) eq "HASH")
    {
#        Carp::carp("Initializing ", $self->fullname, "$name ($retval) to type $type\n");
#        %$retval = ( );
#        bless $retval, $type;
#        $retval->initialize($self->fullname() . $name, $self->_asmunit);
        Carp::croak("Don't know what $name is in import directive");
    }
    Carp::croak("Name refers to a " . ref($retval) . ", not a " . $type)
        unless($retval->isa($type));
    return $retval;
}

sub getNamespaceByName
{
    my $self = shift;
    my $name = shift;
    $self->_namespaces->{$name} or
        $self->_namespaces->{$name} = new Pebble::Namespace($self, $name, $self->_asmunit);
}

sub putNamespaceByName
{
    my $self = shift;
    my $name = shift;
    my $namespace = shift;
    croak("Namespace by name $name already exists in current namespace")
        if $self->_namespaces->{$name};
    croak("Not a namespace!") unless $namespace->isa("Pebble::Namespace");
    $self->_namespaces->{$name} = shift;
}

sub isDeclaredLocally
{
    my $self = shift;
    my $name = shift;
    my $namedthing = $self->getNamedThingByNameLocally($name);
    return $namedthing->isDeclared() if $namedthing;
    return undef;
}

sub declareLocally
{
    my $self = shift;
    my $name = shift;
    my $namedthing = $self->getNamedThingByNameLocally($name);
    unless($namedthing) {
        $self->putByName($name, { }, FALSE);
        $namedthing = $self->getNamedThingByNameLocally($name);
    }
    $namedthing->declare;
    undef;
}

sub debug
{
    my $ns = new Pebble::Namespace(undef);
    $ns->putNamedThingByName("Boo", new Pebble::Arc);
    $ns->putNamedThingByName("Boo", new Pebble::Arc);
}



#
#
#
package Pebble::SpecializedNamespace;
our @ISA = qw( Pebble::Namespace );

sub initialize
{
    my $self = shift;
    $self->SUPER::initialize(@_);
    $self->doSpecialize;
    return $self;
}

sub specialize { croak("Cannot specialize an instance of " . __PACKAGE__); }
sub doSpecialize { }

#
#
#

package Pebble::ExplicitImportNamespace;
our @ISA = qw( Pebble::SpecializedNamespace );

sub getNamedThingByName
{
    my $self = shift;
    my $name = shift;

    return $self->getNamedThingByNameLocally($name);
}

sub doSpecialize { }


#
#
#

package Pebble::ExplicitDeclareNamespace;
our @ISA = qw( Pebble::SpecializedNamespace );

sub doSpecialize { }

#
#
#
package Pebble::FullyExplicitNamespace;
our @ISA = qw( Pebble::SpecializedNamespace );

sub doSpecialize
{
    my $self = shift;
    Pebble::ExplicitDeclareNamespace::doSpecialize($self);
    Pebble::ExplicitImportNamespace::doSpecialize($self);
}

*getNamedThingByName = \&Pebble::ExplicitImportNamespace::getNamedThingByName;

#
#
#
package Pebble::FullyImplicitNamespace;
our @ISA = qw( Pebble::SpecializedNamespace );

sub putByName
{
    my $self = shift;
    my $name = shift;
    my $thing = shift;
    my $declared = 1;
    scalar @_ and $declared = shift;

    $self->SUPER::putByName($name, $thing, $declared);
}

sub doSpecialize
{
    my $self = shift;
    ($_->isDeclared or $_->declare) foreach values %{$self->_kids};
}

sub toString
{
    my $self = shift;
    return $self->fullname;
}

1;
